from unittest import TestCase

class TestModelsResNet2D(TestCase):
    def setUp(self):
        import numpy as np
        from qualia_core.datamodel import RawDataModel
        from qualia_core.datamodel.RawDataModel import RawData, RawDataSets
        train = RawData(np.ones((2, 64, 64, 1), dtype=np.float32), np.array([[1, 0], [1, 0]]))
        test = RawData(np.ones((2, 64, 64, 1), dtype=np.float32), np.array([[1, 0], [0, 1]]))
        self.__data = RawDataModel(RawDataSets(train=train, test=test), name='test_resnet2d')
        self.__model_params = {
            'filters': (4, 6),
            'kernel_sizes': (5, 5),

            'num_blocks': (2,),
            'strides': (1, 2),
            'paddings': (2, 2),

            'prepool': 2,
            'batch_norm': True,
            'bn_momentum': 0.95,

            'dims': 2
            }

    def test_resnet_2d_keras(self):
        from qualia_core import qualia
        from qualia_core.learningframework import Keras
        from qualia_core.learningmodel.keras import ResNet
        from tensorflow.keras.layers import InputLayer, Dense, Activation, Flatten, ZeroPadding2D, Conv2D, AveragePooling2D, MaxPooling2D, BatchNormalization, Dropout, Add
        from tensorflow.keras.activations import relu, softmax

        model = ResNet

        framework = Keras()

        trainresult = qualia.train(self.__data,
                        train_epochs=1,
                        iteration=1,
                        model_name='test_resnet_2d_keras',
                        model=model,
                        model_params=self.__model_params,
                        optimizer={'kind': 'Adam'},
                        framework=framework,
                        )
        self.assertEqual(trainresult.name, 'test_resnet_2d_keras')
        self.assertEqual(trainresult.i, 1)
        self.assertEqual(trainresult.model.input_shape, (None, 64, 64, 1))
        self.assertEqual(trainresult.model.output_shape, (None, 2))

        self.assertEqual(len(trainresult.model.layers), 32)

        self.assertIsInstance(trainresult.model.get_layer('input_1'), InputLayer)
        # prepool
        self.assertIsInstance(trainresult.model.get_layer('average_pooling2d'), AveragePooling2D)
        self.assertEqual(tuple(trainresult.model.get_layer('average_pooling2d').pool_size), (2, 2))
        self.assertEqual(tuple(trainresult.model.get_layer('average_pooling2d').strides), (2, 2))

        # preconv
        self.assertIsInstance(trainresult.model.get_layer('zero_padding2d'), ZeroPadding2D)
        self.assertEqual(tuple(trainresult.model.get_layer('zero_padding2d').padding), ((2, 2), (2, 2)))
        self.assertIsInstance(trainresult.model.get_layer('conv2d'), Conv2D)
        self.assertEqual(tuple(trainresult.model.get_layer('conv2d').kernel.shape), (5, 5, 1, 4)) # 5*5*1*4 = 100
        self.assertEqual(trainresult.model.get_layer('conv2d').bias, None)
        self.assertIsInstance(trainresult.model.get_layer('batch_normalization'), BatchNormalization)
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization').moving_mean.shape), (4,)) # 4
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization').moving_variance.shape), (4, )) # 4
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization').gamma.shape), (4, )) # 4
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization').beta.shape), (4, )) # 4
        self.assertIsInstance(trainresult.model.get_layer('activation'), Activation)
        self.assertEqual(trainresult.model.get_layer('activation').activation, relu)

        # 1st res block
        # 1st conv
        self.assertIsInstance(trainresult.model.get_layer('zero_padding2d_1'), ZeroPadding2D)
        self.assertEqual(tuple(trainresult.model.get_layer('zero_padding2d_1').padding), ((2, 2), (2, 2)))
        self.assertIsInstance(trainresult.model.get_layer('conv2d_1'), Conv2D)
        self.assertEqual(tuple(trainresult.model.get_layer('conv2d_1').kernel.shape), (5, 5, 4, 6)) # 5*4*6 = 600
        self.assertEqual(trainresult.model.get_layer('conv2d_1').bias, None)
        self.assertIsInstance(trainresult.model.get_layer('batch_normalization_1'), BatchNormalization)
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_1').moving_mean.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_1').moving_variance.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_1').gamma.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_1').beta.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model.get_layer('max_pooling2d'), MaxPooling2D)
        self.assertEqual(tuple(trainresult.model.get_layer('max_pooling2d').pool_size), (2, 2))
        self.assertEqual(tuple(trainresult.model.get_layer('max_pooling2d').strides), (2, 2))
        self.assertIsInstance(trainresult.model.get_layer('activation_1'), Activation)
        self.assertEqual(trainresult.model.get_layer('activation_1').activation, relu)
        # 2nd conv
        self.assertIsInstance(trainresult.model.get_layer('conv_ref'), Conv2D)
        self.assertEqual(tuple(trainresult.model.get_layer('conv_ref').kernel.shape), (5, 5, 6, 6)) # 5*6*6 = 900
        self.assertEqual(trainresult.model.get_layer('conv_ref').bias, None)
        self.assertIsInstance(trainresult.model.get_layer('batch_normalization_2'), BatchNormalization)
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_2').moving_mean.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_2').moving_variance.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_2').gamma.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_2').beta.shape), (6, )) # 6
        # shortcut
        self.assertIsInstance(trainresult.model.get_layer('zero_padding2d_2'), ZeroPadding2D)
        self.assertEqual(tuple(trainresult.model.get_layer('zero_padding2d_2').padding), ((2, 2), (2, 2)))
        self.assertIsInstance(trainresult.model.get_layer('conv_shortcut'), Conv2D)
        self.assertEqual(tuple(trainresult.model.get_layer('conv_shortcut').kernel.shape), (1, 1, 4, 6)) # 1*1*4*6 = 24
        self.assertEqual(trainresult.model.get_layer('conv_shortcut').bias, None)
        self.assertIsInstance(trainresult.model.get_layer('batch_normalization_3'), BatchNormalization)
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_3').moving_mean.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_3').moving_variance.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_3').gamma.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_3').beta.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model.get_layer('max_pooling2d_1'), MaxPooling2D)
        self.assertEqual(tuple(trainresult.model.get_layer('max_pooling2d_1').pool_size), (2, 2))
        self.assertEqual(tuple(trainresult.model.get_layer('max_pooling2d_1').strides), (2, 2))
        # add
        self.assertIsInstance(trainresult.model.get_layer('add'), Add)
        self.assertIsInstance(trainresult.model.get_layer('activation_2'), Activation)
        self.assertEqual(trainresult.model.get_layer('activation_2').activation, relu)


        # 2nd conv block
        # 1st conv
        self.assertIsInstance(trainresult.model.get_layer('zero_padding2d_3'), ZeroPadding2D)
        self.assertEqual(tuple(trainresult.model.get_layer('zero_padding2d_3').padding), ((2, 2), (2, 2)))
        self.assertIsInstance(trainresult.model.get_layer('conv2d_2'), Conv2D)
        self.assertEqual(tuple(trainresult.model.get_layer('conv2d_2').kernel.shape), (5, 5, 6, 6)) # 5*6*6 = 900
        self.assertEqual(trainresult.model.get_layer('conv2d_2').bias, None)
        self.assertIsInstance(trainresult.model.get_layer('batch_normalization_4'), BatchNormalization)
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_4').moving_mean.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_4').moving_variance.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_4').gamma.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_4').beta.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model.get_layer('activation_3'), Activation)
        self.assertEqual(trainresult.model.get_layer('activation_3').activation, relu)
        # 2nd conv
        self.assertIsInstance(trainresult.model.get_layer('conv_ref_1'), Conv2D)
        self.assertEqual(tuple(trainresult.model.get_layer('conv_ref_1').kernel.shape), (5, 5, 6, 6)) # 5*6*6 = 900
        self.assertEqual(trainresult.model.get_layer('conv_ref_1').bias, None)
        self.assertIsInstance(trainresult.model.get_layer('batch_normalization_5'), BatchNormalization)
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_5').moving_mean.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_5').moving_variance.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_5').gamma.shape), (6, )) # 6
        self.assertEqual(tuple(trainresult.model.get_layer('batch_normalization_5').beta.shape), (6, )) # 6
        # shortcut
        self.assertIsInstance(trainresult.model.get_layer('zero_padding2d_4'), ZeroPadding2D)
        self.assertEqual(tuple(trainresult.model.get_layer('zero_padding2d_4').padding), ((2, 2), (2, 2)))
        # add
        self.assertIsInstance(trainresult.model.get_layer('add_1'), Add)
        self.assertIsInstance(trainresult.model.get_layer('activation_4'), Activation)
        self.assertEqual(trainresult.model.get_layer('activation_4').activation, relu)

        self.assertIsInstance(trainresult.model.get_layer('max_pooling2d_2'), MaxPooling2D)
        self.assertEqual(tuple(trainresult.model.get_layer('max_pooling2d_2').pool_size), (16, 16))
        self.assertEqual(tuple(trainresult.model.get_layer('max_pooling2d_2').strides), (16, 16))

        self.assertIsInstance(trainresult.model.get_layer('flatten'), Flatten)

        # fc
        self.assertIsInstance(trainresult.model.get_layer('dense'), Dense)
        self.assertEqual(trainresult.model.get_layer('dense').kernel.shape, (6, 2)) # 6*2 = 12
        self.assertEqual(trainresult.model.get_layer('dense').bias.shape, (2, )) # 2

        # softmax
        self.assertIsInstance(trainresult.model.get_layer('activation_5'), Activation)
        self.assertEqual(trainresult.model.get_layer('activation_5').activation, softmax)

        # first layer 10 weights/10 biases, second layer 10*10 weights (10 inputs, 10 outputs)/10 biases, 3rd layer 10*2 weights (10 inputs 2 outputs)/2 biases, 4 bytes (float32)
        self.assertEqual(trainresult.mem_params, (5*5*1*4 + 4*4 + 5*5*4*6 + 6*4 + 5*5*6*6 + 6*4 + 1*1*4*6 + 6*4 + 5*5*6*6 + 6*4 + 5*5*6*6 + 6*4 + 6*2 + 2) * 4)
        self.assertEqual(trainresult.acc, 0.5) # Same data in one or the other class, should have 50% acc
        self.assertEqual(trainresult.framework, framework)

    def test_resnet_2d_pytorch(self):
        from qualia_core import qualia
        from qualia_core.learningframework import PyTorch
        from qualia_core.learningmodel.pytorch import ResNet
        from qualia_core.learningmodel.pytorch.layers import Add
        from torch.nn import Flatten, Linear, ReLU, AvgPool2d, MaxPool2d, Conv2d, BatchNorm2d, Dropout

        model = ResNet

        framework = PyTorch(enable_progress_bar=False)

        trainresult = qualia.train(self.__data,
                        train_epochs=1,
                        iteration=1,
                        model_name='test_resnet_2d_pytorch',
                        model=model,
                        model_params=self.__model_params,
                        optimizer={'kind': 'Adam'},
                        framework=framework,
                        )
        self.assertEqual(trainresult.name, 'test_resnet_2d_pytorch')
        self.assertEqual(trainresult.i, 1)
        #self.assertEqual(trainresult.model.input_shape, (None, 1, 1)) # Not supported in PyTorch
        #self.assertEqual(trainresult.model.output_shape, (None, 2)), Not supported in PyTorch

        self.assertEqual(len(list(trainresult.model.children())), 8)
        self.assertEqual(len(list(trainresult.model.layers.children())), 1)
        self.assertEqual(len(list(trainresult.model.layers[0].children())), 2)
        self.assertEqual(len(list(trainresult.model.layers[0][0].children())), 11)
        self.assertEqual(len(list(trainresult.model.layers[0][1].children())), 7)

        # prepool
        self.assertIsInstance(trainresult.model.prepool, AvgPool2d)
        self.assertEqual(trainresult.model.prepool.kernel_size, 2)
        self.assertEqual(trainresult.model.prepool.stride, 2)

        # preconv
        self.assertIsInstance(trainresult.model.conv1, Conv2d)
        self.assertEqual(tuple(trainresult.model.conv1.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.conv1.weight.shape), (4, 1, 5, 5)) # 4*1*5*5 = 100
        self.assertEqual(trainresult.model.conv1.bias, None)
        self.assertIsInstance(trainresult.model.bn1, BatchNorm2d)
        self.assertEqual(tuple(trainresult.model.bn1.weight.shape), (4,)) # 4
        self.assertEqual(tuple(trainresult.model.bn1.bias.shape), (4, )) # 4
        self.assertIsInstance(trainresult.model.relu1, ReLU)

        # 1st res block
        # 1st conv
        self.assertIsInstance(trainresult.model.layers[0][0].conv1, Conv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv1.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv1.weight.shape), (6, 4, 5, 5)) # 6*4*5*5 = 600
        self.assertEqual(trainresult.model.layers[0][0].conv1.bias, None)
        self.assertIsInstance(trainresult.model.layers[0][0].bn1, BatchNorm2d)
        self.assertEqual(tuple(trainresult.model.layers[0][0].bn1.weight.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.layers[0][0].bn1.bias.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model.layers[0][0].pool1, MaxPool2d)
        self.assertEqual(trainresult.model.layers[0][0].pool1.kernel_size, 2)
        self.assertEqual(trainresult.model.layers[0][0].pool1.stride, 2)
        self.assertIsInstance(trainresult.model.layers[0][0].relu1, ReLU)
        # 2nd conv
        self.assertIsInstance(trainresult.model.layers[0][0].conv2, Conv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv2.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv2.weight.shape), (6, 6, 5, 5)) # 6*6*5 = 900
        self.assertEqual(trainresult.model.layers[0][0].conv2.bias, None)
        self.assertIsInstance(trainresult.model.layers[0][0].bn2, BatchNorm2d)
        self.assertEqual(tuple(trainresult.model.layers[0][0].bn2.weight.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.layers[0][0].bn2.bias.shape), (6, )) # 6
        # shortcut
        self.assertIsInstance(trainresult.model.layers[0][0].sconv, Conv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][0].sconv.padding), (0, 0))
        self.assertEqual(tuple(trainresult.model.layers[0][0].sconv.weight.shape), (6, 4, 1, 1)) # 6*4*1*1 = 24
        self.assertEqual(trainresult.model.layers[0][0].sconv.bias, None)
        self.assertIsInstance(trainresult.model.layers[0][0].sbn, BatchNorm2d)
        self.assertEqual(tuple(trainresult.model.layers[0][0].sbn.weight.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.layers[0][0].sbn.bias.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model.layers[0][0].smax, MaxPool2d)
        self.assertEqual(trainresult.model.layers[0][0].smax.kernel_size, 2)
        self.assertEqual(trainresult.model.layers[0][0].smax.stride, 2)
        # add
        self.assertIsInstance(trainresult.model.layers[0][0].add, Add)
        self.assertIsInstance(trainresult.model.layers[0][0].relu, ReLU)


        # 2nd conv block
        # 1st conv
        self.assertIsInstance(trainresult.model.layers[0][1].conv1, Conv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv1.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv1.weight.shape), (6, 6, 5, 5)) # 6*6*5*5 = 900
        self.assertEqual(trainresult.model.layers[0][1].conv1.bias, None)
        self.assertIsInstance(trainresult.model.layers[0][1].bn1, BatchNorm2d)
        self.assertEqual(tuple(trainresult.model.layers[0][1].bn1.weight.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.layers[0][1].bn1.bias.shape), (6, )) # 6
        self.assertIsInstance(trainresult.model.layers[0][1].relu1, ReLU)
        # 2nd conv
        self.assertIsInstance(trainresult.model.layers[0][1].conv2, Conv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv2.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv2.weight.shape), (6, 6, 5, 5)) # 6*6*5*5 = 900
        self.assertEqual(trainresult.model.layers[0][1].conv2.bias, None)
        self.assertIsInstance(trainresult.model.layers[0][1].bn2, BatchNorm2d)
        self.assertEqual(tuple(trainresult.model.layers[0][1].bn2.weight.shape), (6,)) # 6
        self.assertEqual(tuple(trainresult.model.layers[0][1].bn2.bias.shape), (6, )) # 6
        # add
        self.assertIsInstance(trainresult.model.layers[0][1].add, Add)
        self.assertIsInstance(trainresult.model.layers[0][1].relu, ReLU)

        self.assertIsInstance(trainresult.model.postpool, MaxPool2d)
        self.assertEqual(tuple(trainresult.model.postpool.kernel_size), (16, 16))
        self.assertEqual(tuple(trainresult.model.postpool.stride), (16, 16))

        self.assertIsInstance(trainresult.model.flatten, Flatten)

        # fc
        self.assertIsInstance(trainresult.model.linear, Linear)
        self.assertEqual(trainresult.model.linear.weight.shape, (2, 6)) # 2*6 = 12
        self.assertEqual(trainresult.model.linear.bias.shape, (2, )) # 2


        self.assertEqual(trainresult.mem_params, (5*5*1*4 + 4*2 + 5*5*4*6 + 6*2 + 5*5*6*6 + 6*2 + 1*1*4*6 + 6*2 + 5*5*6*6 + 6*2 + 5*5*6*6 + 6*2 + 6*2 + 2) * 4)
        self.assertEqual(trainresult.acc, 0.5) # Same data in one or the other class, should have 50% acc
        self.assertEqual(trainresult.framework, framework)



#!/usr/bin/env python3

from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import Any

import colorful as cf  # type: ignore[import-untyped]

import qualia_core.utils.args
import qualia_core.utils.config
import qualia_core.utils.plugin
from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.logger import Logger
from qualia_core.utils.merge_dict import merge_dict

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    from qualia_core.postprocessing.Converter import Converter  # noqa: TCH001
    from qualia_core.typing import ConfigDict

logger = logging.getLogger(__name__)

def setup_logging() -> None:
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=logging.INFO)

def qualia(action: str,
           config: ConfigDict,
           configname: str) -> dict[str, list[Any]]:
    # Must be called first to configure devices before initializing Keras
    from qualia_core.utils import TensorFlowInitializer
    tfi = TensorFlowInitializer()
    tfi(seed=config['bench']['seed'],
        reserve_gpu=(config['bench'].get('gpus', None) and 'train' in sys.argv[2:]) or config['bench'].get('reserve_gpu', False),
        gpu_memory_growth=config['bench'].get('gpu_memory_growth', True))

    from qualia_core import command
    from qualia_core.utils import Git

    qualia = qualia_core.utils.plugin.load_plugins(config['bench'].get('plugins', []))

    git = Git()
    Logger.logpath /= config['bench']['name'] # Store logfiles in separate directory according to the config bench name
    Logger.prefix = f'{git.short_hash()}_' # Add prefix according to current git commit
    loggers: dict[str, Logger] = {} # Keep track of the loggers we used to return them

    learningframework = getattr(qualia.learningframework,
                                config['learningframework']['kind'])(**config['learningframework'].get('params', {}))

    dataset = getattr(qualia.dataset, config['dataset']['kind'])(**config['dataset'].get('params', {}))
    converter: type[Converter] = getattr(qualia.converter, config.get('deploy', {}).get('converter', {}).get('kind', ''), None)
    if config.get('deploy', {}).get('converter', '') and not converter:
        logger.error("Converter '%s' not found", config['deploy']['converter'])
        raise ValueError

    deployers: ModuleType = getattr(qualia.deployment, config.get('deploy', {}).get('deployer', {}).get('kind', ''), None)
    if not deployers and converter and converter.deployers is not None:
        deployers = converter.deployers

    dataaugmentations = [getattr(learningframework.dataaugmentations, da['kind'])(**da.get('params', {}))
                             for da in config.get('data_augmentation', {})]

    if action == 'preprocess_data':
        preprocess_data_command = command.PreprocessData()
        loggers.update(preprocess_data_command(qualia=qualia,
                                               dataset=dataset,
                                               config=config))
        return {k: v.content for k, v in loggers.items()}

    dataset_pipeline = dataset
    for preprocessing in config['preprocessing']:
        dataset_pipeline = getattr(qualia.preprocessing,
                                   preprocessing['kind'])(**preprocessing.get('params', {})).import_data(dataset_pipeline)

    data = dataset_pipeline.import_data() if hasattr(dataset_pipeline, 'import_data') else dataset_pipeline()

    if action == 'train':
        train_command = command.Train()
        loggers.update(train_command(qualia=qualia,
                                     learningframework=learningframework,
                                     dataaugmentations=dataaugmentations,
                                     data=data,
                                     config=config,
                                     git=git))
    elif action == 'prepare_deploy':
        prepare_deploy_command = command.PrepareDeploy()
        loggers.update(prepare_deploy_command(qualia=qualia,
                                              learningframework=learningframework,
                                              converter=converter,
                                              deployers=deployers,
                                              data=data,
                                              config=config,
                                              git=git))

    elif action == 'deploy_and_evaluate':
        deploy_and_evaluate_command = command.DeployAndEvaluate()
        loggers.update(deploy_and_evaluate_command(qualia=qualia,
                                                   learningframework=learningframework,
                                                   dataaugmentations=dataaugmentations,
                                                   converter=converter,
                                                   deployers=deployers,
                                                   data=data,
                                                   config=config,
                                                   git=git))
    elif action == 'parameter_research':
        parameter_research = command.ParameterResearch()
        loggers.update(parameter_research(qualia=qualia,
                                          learningframework=learningframework,
                                          dataaugmentations=dataaugmentations,
                                          data=data,
                                          config=config,
                                          git=git))
    else:
        logger.error('Invalid action: %s', action)
        raise ValueError

    return {k: v.content for k, v in loggers.items()}


def main() -> int:
    cf.use_style('solarized')  # type: ignore[untyped-def]

    if len(sys.argv) < 3:
        print(f'Usage: {sys.argv[0]} <config.toml> <preprocess_data|train|prepare_deploy|deploy_and_evaluate|parameter_research> [config_params...]')
        sys.exit(1)

    setup_logging()

    config, configname = qualia_core.utils.config.parse_config(Path(sys.argv[1]))
    config_overwrite = qualia_core.utils.args.parse_args(sys.argv[3:])
    # Overwrite config file params with command line arguments
    config_overwritten = merge_dict(config_overwrite, config, merge_lists=True)
    validated_config = qualia_core.utils.config.validate_config_dict(config_overwritten)
    if validated_config is None:
        logger.error('Could not load configuration.')
        return 1

    loggers = qualia(sys.argv[2], config=validated_config, configname=configname)
    logger.info('%s', loggers)
    return 0

if __name__ == '__main__':
    sys.exit(main())



import numpy as np

from .Preprocessing import Preprocessing

class DatasetSplitterBySubjects(Preprocessing):
    '''Warning: must be applied after Window to get correct split and randomization of windows'''
    def __init__(self, source_subjects: list, dest_subjects: list, source: str='train', dest: str='test'):
        self.__source_subjects = source_subjects
        self.__dest_subjects = dest_subjects
        self.__source = source
        self.__dest = dest

    def __call__(self, datamodel):
        """
            Splits Dataset in two.
            Set dest with 'ratio' percentage of vectors taken from source and set source with '1-ratio' percentage of vectors
            taken from source.
        """
        source = getattr(datamodel.sets, self.__source)
        dest = getattr(datamodel.sets, self.__dest)

        dest = [s for s in source if s.name in self.__dest_subjects]
        source = [s for s in source if s.name in self.__source_subjects]

        setattr(datamodel.sets, self.__dest, dest)
        setattr(datamodel.sets, self.__source, source)

        return datamodel



from dataclasses import dataclass

from .Sensor import Sensor

@dataclass
class GPS(Sensor):
    latitude: float
    longitude: float
    altitude: float
    speed: float
    bearing: float
    accuracy: float



from __future__ import annotations

import sys
from dataclasses import dataclass

from qualia_core.learningmodel.pytorch.Quantizer import QuantizationConfig, Quantizer, update_params
from qualia_core.typing import TYPE_CHECKING

from .Add import Add
from .QuantizedLayer import QuantizedLayer, QuantizerActProtocol

if TYPE_CHECKING:
    import torch  # noqa: TCH002
    from torch import nn  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

@dataclass
class DummyInputQuantizer:
    global_max: None = None

class QuantizedAdd(Add, QuantizerActProtocol, QuantizedLayer):
    def __init__(self,
                 quant_params: QuantizationConfig,
                 activation: nn.Module | None = None) -> None:
        super().__init__()
        self.activation = activation
        quant_params_input = update_params(tensor_type='input', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        self.quantizer_input_a = Quantizer(**quant_params_input)
        self.quantizer_input_b = Quantizer(**quant_params_input)
        self.quantizer_act = Quantizer(**quant_params_act)

    @override
    def forward(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        q_a = self.quantizer_input_a(a)
        q_b = self.quantizer_input_b(b)

        y = super().forward(q_a, q_b)

        if self.activation:
            y = self.activation(y)

        return self.quantizer_act(y)



from __future__ import annotations

import sys

import torch

from .DataAugmentationPyTorch import DataAugmentationPyTorch

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class IntToFloat32(DataAugmentationPyTorch):
    def __init__(self,
                 scale: bool,  # noqa: FBT001
                 evaluate: bool = False,  # noqa: FBT001, FBT002
                 before: bool = False,  # noqa: FBT001, FBT002
                 after: bool = True) -> None:  # noqa: FBT001, FBT002
        super().__init__(evaluate=evaluate, before=before, after=after)
        self.__scale = scale

    def apply(self, x: torch.Tensor, y: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        float_x = x.to(torch.float32)
        if self.__scale:
            iinfo = torch.iinfo(x.dtype)
            return x / iinfo.max, y
        return float_x, y

    @override
    def __call__(self, data: tuple[torch.Tensor, torch.Tensor], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        return self.apply(*data)



import sys
from abc import ABC
from typing import Any, Optional

import torch
from torch import nn

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class CustomBatchNorm(ABC, nn.Module):
    eps: float
    momentum: float
    affine: bool
    track_running_stats: bool
    weight: nn.Parameter
    bias: nn.Parameter
    running_var: nn.Parameter
    running_mean: nn.Parameter
    num_batches_tracked: nn.Parameter
    use_fused_params: bool

    def __init__(self,
                 num_features: int,
                 eps: float = 1e-05,
                 momentum: float = 0.1,
                 affine: bool = True,
                 track_running_stats: bool = True,
                 use_fused_params: bool = True,
                 device: Any = None,
                 dtype: Any = None) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__()

        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.affine = affine
        self.track_running_stats = track_running_stats
        self.use_fused_params = use_fused_params

        if self.affine:
            self.weight = nn.Parameter(torch.ones(num_features, device=device, dtype=dtype))
            self.bias = nn.Parameter(torch.zeros(num_features, device=device, dtype=dtype))

        if track_running_stats:
            self.running_var = nn.Parameter(torch.ones(num_features, device=device, dtype=dtype), requires_grad=False)
            self.running_mean = nn.Parameter(torch.zeros(num_features, device=device, dtype=dtype), requires_grad=False)
            self.num_batches_tracked = nn.Parameter(torch.tensor(0, device=device, dtype=torch.long), requires_grad=False)

    @override
    def extra_repr(self) -> str:
        return (f'{self.num_features}, eps={self.eps}, momentum={self.momentum}, affine={self.affine}, '
                f'track_running_stats={self.track_running_stats}')

    def update_running_stats(self, mean: torch.Tensor, var: torch.Tensor) -> None:
        """Update Exponential Moving Average for input mean and variance."""
        with torch.no_grad():
            self.num_batches_tracked += 1
            self.running_mean *= (1 - self.momentum)
            self.running_mean += self.momentum * mean
            self.running_var *= (1 - self.momentum)
            self.running_var += self.momentum * var

    def compute_stats(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute mean and variance for each channel of input."""
        mean = x.mean(dim=(0, -1))
        var = x.var(dim=(0, -1), correction=0) # PyTorch BatchNorm1d doc unbiased=False, correction=0 is PyTorch 2.0 equivalent
        return mean, var

    def compute_fused_params(self,
                             weight: Optional[torch.Tensor],
                             bias: Optional[torch.Tensor],
                             mean: torch.Tensor,
                             var: torch.Tensor,
                             eps: float) -> tuple[torch.Tensor, torch.Tensor]:
        """Convert the 4(+1) params (weight, bias, mean, variance, +eps) to 2 params to be applied with y = alpha × x + beta.""" # noqa: RUF002
        sigma = (var + eps).sqrt()

        alpha = (weight if weight is not None else 1) / sigma
        beta = (bias if bias is not None else 0) - (weight if weight is not None else 1) * mean / sigma

        return alpha, beta

    def compute_batchnorm_with_fused_params(self,
                                       x: torch.Tensor,
                                       alpha: torch.Tensor,
                                       beta: torch.Tensor) -> torch.Tensor:
        """Compute y = alpha × x + beta.""" # noqa: RUF002
        return x * alpha.unsqueeze(-1) + beta.unsqueeze(-1)

    def compute_batchnorm(self,
                          x: torch.Tensor,
                          weight: Optional[torch.Tensor],
                          bias: Optional[torch.Tensor],
                          mean: torch.Tensor,
                          var: torch.Tensor,
                          eps: float) -> torch.Tensor:
        """Compute batchnorm for input x with given weight, bias, mean, variance and epsilon."""
        if self.use_fused_params:
            alpha, beta = self.compute_fused_params(weight=weight, bias=bias, mean=mean, var=var, eps=eps)
            return self.compute_batchnorm_with_fused_params(x=x, alpha=alpha, beta=beta)

        y = (x - mean.unsqueeze(-1)) / (var.unsqueeze(-1) + eps).sqrt()
        if weight is not None:
            y = y * weight.unsqueeze(-1)
        if bias is not None:
            y = y + bias.unsqueeze(-1)
        return y

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """BatchNorm1d in Python PyTorch.

        Update statistics if required then compute batchnorm for given input using module's own weights, bias and running stats.
        """
        mean, var = self.compute_stats(x=input)

        if self.training and self.track_running_stats:
            self.update_running_stats(mean=mean, var=var)

        if self.training or not self.track_running_stats:
            y = self.compute_batchnorm(x=input,
                                       weight=self.weight if self.affine else None,
                                       bias=self.bias if self.affine else None,
                                       mean=mean,
                                       var=var,
                                       eps=self.eps)
        else:
            y = self.compute_batchnorm(x=input,
                                       weight=self.weight if self.affine else None,
                                       bias=self.bias if self.affine else None,
                                       mean=self.running_mean,
                                       var=self.running_var,
                                       eps=self.eps)
        return y



from unittest import TestCase
from pathlib import Path

class TestGTSRB(TestCase):
    def setUp(self):
        self.path = Path('data')/'GTSRB'

    def test_gtsrb(self):
        import numpy as np
        from qualia_core.dataset import GTSRB

        dataset = GTSRB(path=str(self.path), width=8, height=8)()

        trainY = [int(f.parent.name) for f in (self.path/'Final_Training'/'Images').glob('*/*.ppm')]

        with open(self.path/'GT-final_test.csv') as f:
            next(f) # Skip first line
            testY = np.array([int(l.split(';')[-1]) for l in f.read().splitlines()])

        self.assertEqual(dataset.sets.train.x.shape, (39209, 8, 8, 3))
        self.assertEqual(dataset.sets.test.x.shape, (12630, 8, 8, 3))
        self.assertEqual(dataset.sets.train.y.shape, (39209, ))
        self.assertEqual(dataset.sets.test.y.shape, (12630, ))

        self.assertTrue(np.array_equal(dataset.sets.train.y, trainY))
        self.assertTrue(np.array_equal(dataset.sets.test.y, testY))



from __future__ import annotations

import math
import sys

import torch

from .DataAugmentationPyTorch import DataAugmentationPyTorch

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class TimeWarping(DataAugmentationPyTorch):
    def __init__(self,
                 sigma: float = 0.25,
                 evaluate: bool = False,  # noqa: FBT001, FBT002
                 before: bool = False,  # noqa: FBT001, FBT002
                 after: bool = True) -> None:  # noqa: FBT001, FBT002
        super().__init__(evaluate=evaluate, before=before, after=after)

        self.sigma = sigma

        self.dist = None

    def apply(self, x: torch.Tensor, y: torch.Tensor, device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        if self.dist is None:
            self.dist = torch.distributions.normal.Normal(torch.tensor(1.0, device=device),
                                                          torch.tensor(self.sigma, device=device))

        scale_factor = self.dist.sample()

        # Don't apply time warping if it would result in a null or negative length
        if scale_factor * x.shape[-1] < 1:
            return x, y

        warped_x = torch.nn.functional.interpolate(x,
                                                   scale_factor=(scale_factor, ),
                                                   mode='linear',
                                                   recompute_scale_factor=True)

        warped_x = torch.tile(warped_x, (1, 1, math.ceil(x.shape[-1] / warped_x.shape[-1])))
        warped_x = warped_x[:,:,:x.shape[-1]]

        return warped_x, y

    @override
    def __call__(self, data: tuple[torch.Tensor, torch.Tensor], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        return self.apply(*data, device=device)



from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING, Any

import numpy as np

from qualia_core.datamodel.RawDataModel import RawData, RawDataModel

from .Preprocessing import Preprocessing

if TYPE_CHECKING:
    from qualia_core.dataset.Dataset import Dataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class DatasetSplitter(Preprocessing[RawDataModel, RawDataModel]):
    def __init__(self,
                 source: str = 'train',
                 dest: str = 'test',
                 ratio: float = 0.1) -> None:
        super().__init__()
        self.__source = source
        self.__dest = dest
        self.__ratio = ratio

    @override
    def __call__(self, datamodel: RawDataModel) -> RawDataModel:
        """Split Dataset in two.

        Set dest with 'ratio' percentage of vectors taken from source and set source
        with '1-ratio' percentage of vectors taken from source.
        """
        source: RawData | None = getattr(datamodel.sets, self.__source)
        dest: RawData | None = getattr(datamodel.sets, self.__dest)
        if source is None:
            logger.error('Source set %s does not exist in dataset', self.__source)
            raise ValueError

        perms = np.random.permutation(len(source.x))
        destperms = perms[0:int(len(perms) * self.__ratio)]
        sourceperms = perms[len(destperms):]

        dest_x = source.x[destperms]
        dest_y = source.y[destperms]
        dest_info = source.info[sourceperms] if source.info is not None else None

        if dest is None:
            # Destination set does not exist, create it.
            dest = RawData(x=dest_x, y=dest_y, info=dest_info)
            setattr(datamodel.sets, self.__dest, dest)
        else:
            dest.x = dest_x
            dest.y = dest_y

        source.x = source.x[sourceperms]
        source.y = source.y[sourceperms]
        if source.info is not None:
            source.info = source.info[sourceperms]

        return datamodel

    @override
    def import_data(self, dataset: Dataset[Any]) -> Dataset[Any]:
        # Add dest to list of sets for dataset being loaded
        if self.__dest not in dataset.sets:
            dataset.sets.append(self.__dest)
        return dataset



from __future__ import annotations

import copy
import dataclasses
import logging
import re
import sys
from pathlib import Path
from typing import Any, Optional, Union, cast

import numpy.typing
import tomlkit
from keras.src.engine.node import Node  # type: ignore[import-untyped]
from torch import nn

from qualia_core.typing import TYPE_CHECKING, ModelConfigDict

from .PostProcessing import PostProcessing

if TYPE_CHECKING:
    import keras  #type: ignore[import-untyped] # noqa: TCH002
    from keras.layers import Layer  # type: ignore[import-untyped] # noqa: TCH002

    from qualia_core.learningframework.LearningFramework import LearningFramework  # noqa: TCH001
    from qualia_core.qualia import TrainResult  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

RecursiveDictStr = dict[str, Union['RecursiveDictStr', str]]

class Torch2Keras(PostProcessing[nn.Module]):
    def __init__(self, mapping: str) -> None:
        super().__init__()
        with Path(mapping).open() as f:
            self.__mapping: RecursiveDictStr = tomlkit.parse(f.read())

    def __map_object_to_dict(self,
                             lookup_dict: RecursiveDictStr | str,
                             target_dict: nn.Module | list[nn.Module],
                             result_dict: dict[nn.Module, str] | None = None) -> dict[nn.Module, str]:
        """Get each layer of the PyTorch model according to the hierarchy defined in the 'mapping' parameter and create a flat dict
        with each layer as key and the name of the corresponding Keras layer as value (the leaves of the 'mapping' dict).
        """
        result_dict = result_dict if result_dict is not None else {}

        if not isinstance(lookup_dict, dict):
            if isinstance(target_dict, list):
                logger.error('Expected target_dict to be nn.Module when lookup_dict is str, but got list')
                raise TypeError
            return {target_dict: lookup_dict}

        for k, v in lookup_dict.items():
            if isinstance(target_dict, list):
                result_dict = {**self.__map_object_to_dict(v, target_dict[int(k)], result_dict),
                               **result_dict}
            else:
                target_value = cast(Optional[nn.Module], getattr(target_dict, k, None))

                if target_value is None:
                    logger.warning('Could not find %s in target_dict, ignoring', k)
                    continue

                result_dict = {
                        **self.__map_object_to_dict(v, target_value, result_dict),
                        **result_dict}
        return result_dict

    def __extract_weight_type_from_name(self, name: str) -> str | None:
        wt = re.search('^.+?/(.*):.+?$', name)
        if wt is None:
            return None
        return wt.group(1)

    def __reformat_fc_weights_data(self, tf_layers: list[Layer]) -> None:
        # After Flatten comes Dense, must reshape weights and swap axes for channels_first
        from tensorflow.keras.layers import Dense, Flatten  # type: ignore[import-untyped]
        for layer in tf_layers:
            if isinstance(layer, Flatten):
                for outnode in cast(list[Node], layer.outbound_nodes):
                    dense = outnode.layer
                    if not isinstance(dense, Dense):
                        logger.error('Flatten layer must be followed by Dense')
                        raise TypeError

                    kernel = cast(numpy.typing.NDArray[Any], dense.kernel.numpy())
                    input_shape = cast(tuple[int, ...], layer.input_shape)
                    units = cast(int, dense.units)

                    # reshape using Flatten input shape (for example last Conv output)
                    kernel = kernel.reshape(input_shape[-1:] + input_shape[1:-1] + (units, ))
                    kernel = kernel.swapaxes(0, 1)
                    kernel = kernel.reshape((-1, units))
                    dense.set_weights([kernel] + dense.get_weights()[1:])

    @override
    def __call__(self,
                 trainresult: TrainResult,
                 model_conf: ModelConfigDict) -> tuple[TrainResult, ModelConfigDict]:
        from keras.layers import Conv1D, Dense  # type: ignore[import-untyped]

        import qualia_core.learningmodel.keras as keras_learningmodels
        from qualia_core.learningframework import Keras

        framework = trainresult.framework
        model = trainresult.model
        model_nparams = framework.n_params(model)
        model.eval()

        # For activations_range, convert module name to identifier, '.' replaced with '_'
        # This is what QuantizationAwareTraining activations_range.txt uses
        from torch.fx.graph import _Namespace, _snake_case
        namespace = _Namespace()
        name_to_layer_torch = {namespace.create_name(_snake_case(name), m)
                                if name else name: m for name, m in model.named_modules()}

        framework = Keras()
        tf_params = copy.deepcopy(model_conf.get('params', {}))

        if 'quant_params' in tf_params:
            logger.warning('Dropping quant_params from model params, construction of quantized model is unsupported for Keras')
            del tf_params['quant_params']

        tf_model = getattr(keras_learningmodels, model_conf['kind'])(**tf_params)
        tf_model_nparams = framework.n_params(tf_model)

        framework.summary(tf_model)

        if model_nparams != tf_model_nparams:
            logger.error('Different number of parameters: PyTorch=%s, Keras=%s',
                         model_nparams,
                         tf_model_nparams)
            raise RuntimeError

        weights_name_tf_to_torch = { 'kernel': 'weight', 'bias': 'bias' }
        weights_transpose = {
            Conv1D: { 'kernel': (2, 1, 0), 'bias': (0,) },
            Dense: {'kernel': (1, 0), 'bias': (0,)},
        }

        layers_torch_to_tf = self.__map_object_to_dict(self.__mapping, model)
        layers_tf_to_torch = {v: k for k, v in layers_torch_to_tf.items()} # Reverse lookup

        for layer in tf_model.layers:
            if len(layer.weights) > 0: #layer need weights
                for weight in layer.weights:
                    weight_type = self.__extract_weight_type_from_name(weight.name)
                    if weight_type is None:
                        logger.error("Could not find weight type in name '%s'", weight.name)
                        raise RuntimeError
                    torch_layer = layers_tf_to_torch[layer.name]
                    weight_values_torch = getattr(torch_layer, weights_name_tf_to_torch[weight_type])
                    weight_values_np = weight_values_torch.data.cpu().numpy()

                    weight.assign(weight_values_np.transpose(weights_transpose[layer.__class__][weight_type]))

        self.__reformat_fc_weights_data(tf_model.layers)

        # Compile Keras model
        tf_model.compile(loss='categorical_crossentropy', metrics=['categorical_accuracy'])

        ### Convert Activations Range file with TF layer names
        activations_range_filepath = Path('out')/'learningmodel'/f'{trainresult.name}_r{trainresult.i}_activations_range.txt'
        if activations_range_filepath.is_file():
            activations_range: dict[str, str] = {}
            with activations_range_filepath.open() as f:
                for line in f:
                    r = line.strip().split(',', 1)
                    activations_range[r[0]] = r[1]
            with activations_range_filepath.with_suffix('.h5.txt').open('w') as f:
                for lname, arange in activations_range.items():
                    layer_torch = name_to_layer_torch[lname]
                    if layer_torch not in layers_torch_to_tf:
                        logger.warning('Unusued layer %s', lname)
                    else:
                        print(f'{layers_torch_to_tf[layer_torch]},{arange}', file=f)
        else:
            logger.info('No activation range file found, skipping conversion')

        new_model_conf = copy.deepcopy(model_conf)
        new_model_conf['params'] = tf_params

        return dataclasses.replace(trainresult, model=tf_model, framework=framework, log=False), new_model_conf

    @override
    def process_framework(self, framework: LearningFramework[nn.Module]) -> LearningFramework[keras.Model]:
        from qualia_core.learningframework.Keras import Keras
        return Keras()



from dataclasses import asdict, astuple, dataclass, fields
from typing import Any, NamedTuple


class StatsFields(NamedTuple):
    name: str
    i: int
    quantization: str
    params: int
    mem_params: int
    accuracy: float # between 0 and 1
    avg_time: float # in seconds
    rom_size: int # in bytes
    ram_size: int # in bytes

@dataclass
class Stats:
    name: str = ''
    i: int = -1
    quantization: str = 'float32'
    params: int = -1
    mem_params: int = -1
    accuracy: float = -1 # between 0 and 1
    avg_time: float = -1 # in seconds
    rom_size: int = -1 # in bytes
    ram_size: int = -1 # in bytes

    @classmethod
    def fieldnames(cls) -> tuple[str, ...]:
        return tuple(f.name for f in fields(Stats))

    def asdict(self) -> dict[str, Any]:
        return asdict(self)

    def astuple(self) -> tuple[Any, ...]:
        return astuple(self)

    def asnamedtuple(self) -> StatsFields:
        return StatsFields(**asdict(self))



from pathlib import Path
import re
import sys
from qualia_core.utils.process import subprocesstee                                                                                     
from qualia_core.evaluation.Stats import Stats

import qualia_core.evaluation.STM32CubeAI as STM32CubeAIBase

class STM32CubeAI(STM32CubeAIBase):
    def __init__(self,
        dev: str='/dev/ttyACM0',
        baudrate: int=921600):
        super().__init__(mode='stm32', stm32cubeai_args=('--desc',f'{dev}:{baudrate}'))



from __future__ import annotations

import copy
import dataclasses
import logging
import sys
from typing import TYPE_CHECKING, Any

from torch import nn

from qualia_core.learningmodel.pytorch.layers import layers as custom_layers

from .PostProcessing import PostProcessing

if TYPE_CHECKING:
    from torch.fx.graph import Graph
    from torch.fx.graph_module import GraphModule

    from qualia_core.qualia import TrainResult
    from qualia_core.typing import ModelConfigDict

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)


class FuseBatchNorm(PostProcessing[nn.Module]):
    custom_layers: tuple[type[nn.Module], ...] = custom_layers

    patterns: list[tuple[type[nn.Module], type[nn.Module]]]

    def __init__(self) -> None:
        super().__init__()

        from torch import nn

        self.patterns = [
                (nn.Conv1d, nn.BatchNorm1d),
                (nn.Conv2d, nn.BatchNorm2d),
                (nn.Conv3d, nn.BatchNorm3d),
                ]

    # Copied from torch/fx/experimental/optimization.py because we need a custom Tracer for our custom layers
    def fuse(self, model: nn.Module, inplace: bool = False) -> GraphModule:
        """Fuses convolution/BN layers for inference purposes.

        Will deepcopy your model by default, but can modify the model inplace as well.
        """
        from torch.fx._symbolic_trace import Tracer
        from torch.fx.experimental.optimization import matches_module_pattern, replace_node_module
        from torch.fx.graph_module import GraphModule
        from torch.nn.utils.fusion import fuse_conv_bn_eval

        # Custom tracer that generates call_module for our custom Qualia layers instead of attempting to trace their forward()
        class TracerCustomLayers(Tracer):
            def __init__(self, custom_layers: tuple[type[nn.Module], ...], *args: Any, **kwargs: Any) -> None:
                super().__init__(*args, **kwargs)
                self.custom_layers = custom_layers

            @override
            def is_leaf_module(self, m: nn.Module, module_qualified_name : str) -> bool:
                return super().is_leaf_module(m, module_qualified_name) or isinstance(m, self.custom_layers)


        if not inplace:
            model = copy.deepcopy(model)

        tracer = TracerCustomLayers(custom_layers=self.custom_layers)
        graph = tracer.trace(model)
        graph.print_tabular()
        graphmodule = GraphModule(tracer.root, graph, tracer.root.__class__.__name__)
        fx_model = graphmodule

        modules: dict[str, nn.Module] = dict(fx_model.named_modules())
        new_graph: Graph = copy.deepcopy(fx_model.graph)

        for pattern in self.patterns:
            for node in new_graph.nodes:
                if matches_module_pattern(pattern, node, modules):
                    if len(node.args[0].users) > 1:  # Output of conv is used by other nodes
                        continue
                    conv = modules[node.args[0].target]
                    bn = modules[node.target]
                    if not bn.track_running_stats:
                        continue
                    fused_conv = fuse_conv_bn_eval(conv, bn)
                    replace_node_module(node.args[0], modules, fused_conv)
                    node.replace_all_uses_with(node.args[0])
                    new_graph.erase_node(node)
        return GraphModule(fx_model, new_graph)

    @override
    def __call__(self, trainresult: TrainResult, model_conf: ModelConfigDict) -> tuple[TrainResult, ModelConfigDict]:
        model = trainresult.model
        model.eval() # Can only fuse models in eval mode

        fused_model = self.fuse(model, inplace=True)

        # Copy input_shape/output_shape into generated model
        fused_model.input_shape = model.input_shape
        fused_model.output_shape = model.output_shape

        model_name = f'{trainresult.name}_fused'

        logger.info('Model after BatchNorm/Conv fusion')
        trainresult.framework.summary(fused_model)

        logger.info('Evaluation on test dataset after BatchNorm/Conv fusion')
        metrics = trainresult.framework.evaluate(fused_model,
                                                 trainresult.testset,
                                                 batch_size=trainresult.batch_size,
                                                 dataaugmentations=trainresult.dataaugmentations,
                                                 experimenttracking=trainresult.experimenttracking,
                                                 dataset_type='test',
                                                 name=model_name)

        new_model_conf = copy.deepcopy(model_conf)
        new_model_conf.get('params', {})['batch_norm'] = False

        return (dataclasses.replace(trainresult, name=model_name, model=fused_model, acc=metrics.get('testacc', None)),
                new_model_conf)

    @override
    def process_model(self, model: nn.Module, model_conf: ModelConfigDict) -> tuple[nn.Module, ModelConfigDict]:
        model = model.eval() # Can only fuse models in eval mode
        fused_model = self.fuse(model, inplace=True)
        # Copy input_shape/output_shape into generated model
        fused_model.input_shape = model.input_shape
        fused_model.output_shape = model.output_shape
        model_conf.get('params', {})['batch_norm'] = False
        return fused_model, model_conf

    @override
    def process_name(self, name: str) -> str:
        return f'{name}_fused'



from unittest import TestCase
from pathlib import Path

class TestWSMNIST(TestCase):
    def setUp(self):
        self.path = Path('data')/'WSMNIST'

    def test_wsmnist_spoken(self):
        import numpy as np
        from qualia_core.dataset import WSMNIST

        dataset = WSMNIST(path=str(self.path), variant='spoken')()

        trainX = np.load(self.path/'data_sp_train.npy')
        trainY = np.load(self.path/'labels_train.npy')
        testX = np.load(self.path/'data_sp_test.npy')
        testY = np.load(self.path/'labels_test.npy')
        
        self.assertEqual(dataset.sets.train.x.shape, (60000, 39, 13))
        self.assertEqual(dataset.sets.test.x.shape, (10000, 39, 13))
        self.assertEqual(dataset.sets.train.y.shape, (60000, ))
        self.assertEqual(dataset.sets.test.y.shape, (10000, ))

        self.assertTrue(np.allclose(dataset.sets.train.x.reshape((60000, -1)), trainX))
        self.assertTrue(np.allclose(dataset.sets.test.x.reshape((10000, -1)), testX))
        self.assertTrue(np.allclose(dataset.sets.train.y, trainY))
        self.assertTrue(np.allclose(dataset.sets.test.y, testY))

    def test_wsmnist_written(self):
        import numpy as np
        from qualia_core.dataset import WSMNIST

        dataset = WSMNIST(path=str(self.path), variant='written')()

        trainX = np.load(self.path/'data_wr_train.npy')
        trainY = np.load(self.path/'labels_train.npy')
        testX = np.load(self.path/'data_wr_test.npy')
        testY = np.load(self.path/'labels_test.npy')
        
        self.assertEqual(dataset.sets.train.x.shape, (60000, 28, 28, 1))
        self.assertEqual(dataset.sets.test.x.shape, (10000, 28, 28, 1))
        self.assertEqual(dataset.sets.train.y.shape, (60000, ))
        self.assertEqual(dataset.sets.test.y.shape, (10000, ))

        self.assertTrue(np.allclose(dataset.sets.train.x.reshape((60000, -1)), trainX))
        self.assertTrue(np.allclose(dataset.sets.test.x.reshape((10000, -1)), testX))
        self.assertTrue(np.allclose(dataset.sets.train.y, trainY))
        self.assertTrue(np.allclose(dataset.sets.test.y, testY))



from abc import ABC, abstractmethod
from typing import Optional

from qualia_core.typing import RecursiveConfigDict


class ExperimentTracking(ABC):
    @abstractmethod
    def start(self, name: Optional[str] = None) -> None:
        ...

    @abstractmethod
    def stop(self) -> None:
        ...

    @classmethod
    def initializer(cls) -> None:
        pass

    @abstractmethod
    def _hyperparameters(self, params: RecursiveConfigDict) -> None:
        ...

    # Lambda used to obtain concrete property from abstract setter
    hyperparameters = property(None, lambda self, x: self._hyperparameters(x))



from __future__ import annotations

import itertools
from datetime import datetime
from pathlib import Path
from typing import Any

import colorful as cf  # type: ignore[import-untyped]

from qualia_core.evaluation.Stats import StatsFields
from qualia_core.qualia import deploy, evaluate, gen_tag, instantiate_model
from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.logger import CSVLogger, Logger

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation  # noqa: TCH001
    from qualia_core.datamodel.RawDataModel import RawDataModel  # noqa: TCH001
    from qualia_core.learningframework.LearningFramework import LearningFramework  # noqa: TCH001
    from qualia_core.postprocessing.Converter import Converter  # noqa: TCH001
    from qualia_core.typing import ConfigDict
    from qualia_core.utils import Git  # noqa: TCH001
    from qualia_core.utils.plugin import QualiaComponent  # noqa: TCH001


class DeployAndEvaluate:
    def __call__(self,  # noqa: PLR0913
                 qualia: QualiaComponent,
                 learningframework: LearningFramework[Any],
                 dataaugmentations: list[DataAugmentation],
                 converter: type[Converter],
                 deployers: ModuleType,
                 data: RawDataModel,
                 config: ConfigDict,
                 git: Git) -> dict[str, Logger[StatsFields]]:
        loggers: dict[str, Logger[StatsFields]] = {}
        log: CSVLogger[StatsFields] = CSVLogger('evaluate', file=Path('evaluate')/f'{git.short_hash()}_{datetime.now():%Y-%m-%d_%H-%M-%S}.csv')
        loggers['evaluate'] = log
        # Write column names
        log.fields = StatsFields

        for i in range(config['bench']['first_run'], config['bench']['last_run']+1):
            for m, q, o, c in itertools.product(config['model'],
                                                config['deploy']['quantize'],
                                                config['deploy']['optimize'],
                                                config['deploy'].get('compress', [1])):
                if m.get('disabled', False):
                    continue

                # Postprocessings can change model name, frameworks, mem_params
                model_name = m['name']
                fmem_params = lambda framework, model: framework.n_params(model) * 4 # By default models have 4-bytes params
                for postprocessing in config.get('postprocessing', []):
                    ppp = {k: v for k,v in postprocessing.get('params', {}).items()} # Workaround tomlkit bug where some nested dict would lose their items
                    pp = getattr(qualia.postprocessing, postprocessing['kind'])(**ppp)
                    model_name = pp.process_name(model_name)
                    learningframework = pp.process_framework(learningframework)
                    #mem_params = pp.process_mem_params(mem_params) #FIXME: handled by converter instead

                #model = framework.load(f'{model_name}_r{i}', m['kind'])


                # Instantiate model
                model = instantiate_model(dataset=data.sets.test,
                                          framework=learningframework,
                                          model=getattr(learningframework.learningmodels, m['kind']),
                                          model_params=m.get('params', {}),
                                          model_name=model_name,
                                          iteration=i,
                                          load=False # Model params will be loaded after postprocessings
                                          )

                # Postprocessings can change model topology with PyTorch, needs to be done after instantiating model with new name
                for postprocessing in config.get('postprocessing', []):
                    pp = getattr(qualia.postprocessing, postprocessing['kind'])(**postprocessing.get('params', {}))
                    model, m = pp.process_model(model, m)

                # Show model architecture
                learningframework.summary(model)

                # Load weights after topology optionally changed
                model = learningframework.load(f'{model_name}_r{i}', model)

                # Converter can affect mem_params if there is quantization
                if not converter:
                    converter = getattr(m['kind'], 'converter', None)
                fmem_params = converter(quantize=q).process_mem_params(fmem_params) if converter else fmem_params

                tag = gen_tag(model_name, q, o, i, c)
                print(f'{cf.bold}Deploying {cf.blue}{model_name}{cf.close_fg_color}, run {cf.red}{i}{cf.close_fg_color}, tag {cf.yellow}{tag}{cf.reset}')
                result_deploy = deploy(model_kind=m['kind'],
                                    deploy_target=config['deploy']['target'],
                                    tag=tag,
                                    deployers=deployers,
                                    deployer_params=config['deploy'].get('deployer', {}).get('params', {}))
                if not result_deploy:
                    continue

                print(f'{cf.bold}Evaluating {cf.blue}{model_name}{cf.close_fg_color}, run {cf.red}{i}{cf.close_fg_color}, tag {cf.yellow}{tag}{cf.reset}')
                result = evaluate(datamodel=data,
                                model_kind=m['kind'],
                                model_name=model_name,
                                model=model,
                                framework=learningframework,
                                iteration=i,
                                target=config['deploy']['target'],
                                quantization=q,
                                fmem_params=fmem_params,
                                tag=tag,
                                limit=config['deploy'].get('limit', None),
                                evaluator=getattr(result_deploy,
                                                   'evaluator',
                                                   getattr(converter, 'evaluator', None) if converter else None),
                                evaluator_params=config['deploy'].get('evaluator', {}).get('params', {}),
                                dataaugmentations=dataaugmentations)

                # fill in ROM and RAM size from deployment result
                if result:
                    result.rom_size = result_deploy.rom_size
                    result.ram_size = result_deploy.ram_size

                print(result)
                if result is not None:
                    log(result.asnamedtuple())
        return loggers



from __future__ import annotations

import logging
import sys
from typing import Literal, cast, overload

import torch
from torch import nn

from qualia_core.typing import QuantizationConfigDict, QuantizerConfigDict
from qualia_core.typing import QuantizationConfigDict as QuantizationConfig
from qualia_core.typing import QuantizerConfigDict as QuantizerConfig

from .grad_functions import IntNoGradientceil, IntNoGradientfloor, IntNoGradientround
from .range_setting import MSE_analysis_range_setting, MSE_simulation_range_setting, minmax_range_setting

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

def grad_scale(x: torch.Tensor, scale: torch.Tensor) -> torch.Tensor:
    y = x
    y_grad = x * scale
    return (y - y_grad).detach() + y_grad


def round_pass(x: torch.Tensor) -> torch.Tensor:
    y = x.round()
    y_grad = x
    return (y - y_grad).detach() + y_grad

def maxWeight(maxi: torch.Tensor) -> int:
        maxi = torch.tensor(maxi, device=maxi.device)
        if maxi == 0:
            maxi = torch.tensor(-1, device=maxi.device)
        elif abs(maxi) != float('inf'):
            maxi = -(torch.log2(maxi).floor().int() + 1)
        else:
            maxi = torch.tensor(0, device=maxi.device)
        return cast(int, maxi.item()) # maxi is necessarily an int in all 3 cases

def flat_and_cat(w: torch.Tensor, bias_tensor: torch.Tensor | None) -> torch.Tensor:
    w_and_biases = torch.flatten(w)
    if bias_tensor is not None:
        w_and_biases = torch.cat((w_and_biases, torch.flatten(bias_tensor)))
    return w_and_biases

class Quantizer(nn.Module):
    global_min: torch.Tensor
    global_max: torch.Tensor

    def __init__(self,  # noqa: PLR0913
                 quant_enable: bool = True,  # noqa: FBT001, FBT002
                 LSQ: bool = False,  # noqa: FBT001, FBT002
                 bits: int = 8,
                 force_q: int | None = None,
                 range_setting: str = 'minmax',
                 roundtype: str = 'nearest',
                 tensor_type: str | None = None,
                 quantype: str | None = None,
                 is_asymmetric: bool = False) -> None:  # noqa: FBT001, FBT002
        super().__init__()

        self.bits = bits
        self.quantype = quantype
        self.range_setting = range_setting
        self.is_asymmetric = is_asymmetric
        self.LSQ = LSQ
        self.force_q = force_q
        self.roundtype = roundtype
        self.tensor_type = tensor_type
        self.quant_enable = quant_enable
        self.global_max =  torch.tensor(-float('inf'))
        self.global_min =  torch.tensor(+float('inf'))

        # unsigned activation might be used later
        #   self.neg_bits = 0
        #   self.pos_bits = 2 ** self.bits - 1

        # Set LSQscale weight to None, it will be created later
        self.LSQscale = None

        if self.is_asymmetric:
            # asymmetricsigned weight/activation is quantized to [-2^(b-1), 2^(b-1)-1]
            self.neg_bits = - 2 ** (self.bits - 1)
            self.pos_bits = 2 ** (self.bits - 1) - 1
        else:
            # symmetric signed weight/activation is quantized to [-2^(b-1)+1, 2^(b-1)-1]
            self.neg_bits = - 2 ** (self.bits - 1) + 1
            self.pos_bits = 2 ** (self.bits - 1) - 1

    @override
    def forward(self,
                w: torch.Tensor,
                bias_tensor: torch.Tensor | None = None) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
        # Override the forward method from the base nn.Module class.

        # If quantization is off then return input w +/- bias_tensor
        if not self.quant_enable:
            return w if bias_tensor is None else (w, bias_tensor)

        # Compute/update self.global_max and self.global_min
        self.update_min_and_max(w, bias_tensor)



        ###########################
        ########### LSQ ###########
        if self.LSQ:
            ############### LSQ ###############
            w, bias_tensor, scale, offset = self.applyLSQ(w, bias_tensor)
            ###################################

        else:
        ########### PTQ and QAT ###########
            scale, offset = self.compute_scale_and_offset(flat_and_cat(w, bias_tensor), self.global_max, self.global_min)

            if bias_tensor is None:
                w = self.quant_dequant( w, scale, offset)
            else:
                w, bias_tensor = self.quant_dequant( w, scale, offset), self.quant_dequant( bias_tensor, scale, offset)

        ###################################

        # Return quantized w +/- bias_tensor
        return w if bias_tensor is None else (w, bias_tensor)

    @overload
    def __call__(self,
                 w: torch.Tensor,
                 bias_tensor: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        ...

    @overload
    def __call__(self,
                 w: torch.Tensor) -> torch.Tensor:
        ...

    @override
    def __call__(self,
                 w: torch.Tensor,
                 bias_tensor: torch.Tensor | None = None) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
        return super().__call__(w, bias_tensor=bias_tensor)


    def applyLSQ(self,
                 w: torch.Tensor,
                 bias_tensor: torch.Tensor | None = None)-> tuple[torch.Tensor, torch.Tensor | None, torch.Tensor, torch.Tensor]:
        if self.LSQscale is None:
            w_and_biases = flat_and_cat(w, bias_tensor)
            self.LSQscale = (torch.nn.Parameter(w_and_biases.detach().abs().mean() * 2 / (self.pos_bits ** 0.5))
                             or torch.nn.Parameter(torch.ones(1, device=w.device)))

        if not self.is_asymmetric:
            if self.training:
                # Original here https://arxiv.org/pdf/1902.08153.pdf 2.2 STEP SIZE GRADIENT SCALE
                # s_grad_scale = 1.0 / ((torch.tensor(self.pos_bits, device=w.device) * w.size(1)) ** 0.5) if ("act" in self.tensor_type) or ("input" in self.tensor_type) else 1.0 / ((torch.tensor(self.pos_bits, device=w.device) * w.numel()) ** 0.5)
                # But choosen from  here https://github.com/zhutmost/lsq-net/blob/master/quan/quantizer/lsq.py
                s_grad_scale = 1.0 / ((torch.tensor(self.pos_bits, device=w.device) * w.numel()) ** 0.5)
                scale = grad_scale(self.LSQscale, s_grad_scale)
            else:
                scale = self.LSQscale

            # If scale at 0 put 1
            scale = scale or torch.tensor(1, device=w.device)
            offset = torch.tensor(0, device=w.device)

            # Apply scale and clamp
            w = torch.clamp(w / scale, torch.tensor(self.neg_bits, device=w.device), torch.tensor(self.pos_bits, device=w.device))
            # Apply round and dequantize
            w = round_pass(w) * scale
            if bias_tensor is not None:
                bias_tensor = torch.clamp(bias_tensor / scale,
                                          torch.tensor(self.neg_bits, device=w.device),
                                          torch.tensor(self.pos_bits, device=w.device))
                bias_tensor = round_pass(bias_tensor) * scale
        else:
            raise ValueError('Asym LSQ+ not implemented.')
        return w, bias_tensor, scale, offset

    def update_min_and_max(self, w: torch.Tensor, bias_tensor: torch.Tensor | None = None) -> None:
        # Update only at training
        if self.training:
                # Concatenate weights and biases to be able to compute max and min
                flat_w = flat_and_cat(w, bias_tensor)
                if self.range_setting == 'minmax':
                    max_val, min_val = minmax_range_setting(flat_w, self.is_asymmetric)
                elif self.range_setting == 'MSE_simulation':
                    max_val, min_val = MSE_simulation_range_setting(flat_w, self.bits, self.is_asymmetric)
                elif self.range_setting == 'MSE_analysis':
                    max_val, min_val = MSE_analysis_range_setting(flat_w, self.bits, self.is_asymmetric)
                else:
                    raise ValueError(self.range_setting, 'not implemented.')
                self.global_max = max_val if self.tensor_type in ['w', 'bias'] else torch.max(self.global_max, max_val).detach()
                self.global_min = min_val if self.tensor_type in ['w', 'bias'] else torch.min(self.global_min, min_val).detach()

    def compute_scale_and_offset(self,
                                 w: torch.Tensor,
                                 max_val: torch.Tensor,
                                 min_val: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        max_val = torch.tensor(max_val, device=w.device)
        min_val = torch.tensor(min_val, device=w.device)
        scale = torch.tensor(1, device=w.device)
        offset = torch.tensor(0, device=w.device)
        if max_val != 0 :
            if self.force_q is not None:
                    scale = 1/torch.pow(2.0, torch.tensor(self.force_q, device=w.device))
            elif self.quantype == 'fxp' and not self.is_asymmetric:
                    left_bw = -(torch.log2(max_val).floor().int() + 1)
                    scale = 1/torch.pow(2.0, self.bits - 1 + left_bw)
            elif self.quantype == 'fake':
                max_range = max_val - min_val
                bits_range = float(self.pos_bits - self.neg_bits)
                scale = (max_range) / (bits_range)
            else:
                raise ValueError(self.quantype, 'not implemented. - Does not work with asymmetric.')

        scale = torch.tensor(1, device=w.device) if scale == torch.tensor(0, device=w.device) else scale
        offset = (torch.clamp((-torch.round(min_val / scale) + self.neg_bits), min_val, max_val)
                  if self.is_asymmetric
                  else torch.tensor(0, device=w.device))

        return scale, offset

    def quant_dequant(self, x_f: torch.Tensor, scale: torch.Tensor, offset: torch.Tensor) -> torch.Tensor:
        x_int = x_f / scale

        if self.roundtype == 'nearest':
            x_int = IntNoGradientround.apply(x_int)
        elif self.roundtype == 'floor':
            x_int = IntNoGradientfloor.apply(x_int)
        elif self.roundtype == 'ceil':
            x_int = IntNoGradientceil.apply(x_int)
        else:
            raise ValueError(self.roundtype, 'not implemented.')

        if self.is_asymmetric:
            x_int += offset

        x_q = torch.clamp(x_int, torch.tensor(self.neg_bits, device=x_f.device), torch.tensor(self.pos_bits, device=x_f.device))

        if self.is_asymmetric:
            x_q -= offset

        return x_q * scale

    @property
    def fractional_bits(self) -> int | None:
        if not self.quant_enable:
            logger.info('Quantizer for tensor type %s disabled, no fractional bits to return', self.tensor_type)
            return None

        if self.quantype != 'fxp':
            logger.warning("Quantizer type '%s' is not fixed-point, no fractional bits to return", self.quantype)
            return None

        if self.is_asymmetric:
            logger.warning('Asymmetric quantization unsupported, no fractional bits to return')
            return None

        if self.force_q is not None:
            return self.force_q

        return self.bits - 1 + maxWeight(self.global_max)


def update_params(tensor_type: Literal['act', 'input', 'w', 'bias'],
                  quant_params: QuantizationConfigDict) -> QuantizerConfigDict:
    # Create a copy of quant_params
    quant_params_copy = quant_params.copy()

    # Remove all tensor_type keys from tensor_params
    _ = quant_params_copy.pop('act', None)
    _ = quant_params_copy.pop('input', None)
    _ = quant_params_copy.pop('w', None)
    _ = quant_params_copy.pop('bias', None)

    tensor_params = QuantizerConfigDict(quant_params_copy)

    new_params = quant_params.get(tensor_type, None)

    # If tensor_type exists as a key in tensor_params, update tensor_params with the dict associated to the tensor_type key
    if new_params is not None:
        tensor_params.update(new_params)

    # Modify the "tensor_type" key by appending "_type" to keep tensor_type information
    tensor_params['tensor_type'] = tensor_type + '_type'

    # Return the updated tensor_params dictionary
    return tensor_params

__all__ = ['QuantizerConfig', 'QuantizationConfig']



from __future__ import annotations

import logging
from typing import Any, Callable

from qualia_core.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from qualia_core.datamodel.DataModel import DataModel  # noqa: TCH001
    from qualia_core.typing import ConfigDict
    from qualia_core.utils.logger import Logger  # noqa: TCH001
    from qualia_core.utils.plugin import QualiaComponent  # noqa: TCH001

logger = logging.getLogger(__name__)

class PreprocessData:
    def __call__(self,
                 qualia: QualiaComponent,
                 dataset: Callable[[], DataModel[Any]],
                 config: ConfigDict) -> dict[str, Logger]:

        data = dataset()
        for preprocessing in config['preprocessing']:
            data = getattr(qualia.preprocessing, preprocessing['kind'])(**preprocessing.get('params', {}))(data)
        logger.info('Exporting data')
        _ = data.export()

        return {}



from .PostProcessing import PostProcessing

import math
import torch
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from pathlib import Path
import pickle
import os
import re
from os import listdir
from os.path import isfile, join
import numpy as np

class Distribution(PostProcessing):
    def __init__(self, model: dict=None, method: str="layer_wise", bins: int=10, min: float=-5., max: float=5., output_layer:bool=True, output_pdf:bool=False):
        """ Generate a PDF with the distribution of all layers

            :param method: ["layer_wise", "network_wise"] Define the way to save the model parameters
            :param bins: Bins for histogram
            :param min: min for histogram border
            :param max: max for histogram border
            :param output_layer: generate .pkl with layer informations inside. Required for the first use.
            :param output_pdf: generate a pdf with distribution per {use method}
            :return trainresult:
        """
        print(f"{method=}")
        self.params  = {}
        self.bins = bins
        self.min    = min
        self.max    = max
        self.output_layer   = output_layer
        self.output_pdf     = output_pdf

        self.output_path    = Path('out')/'distribution'
        self.model_name     = ""

        if method in ["layer_wise", "network_wise"]:
            self.method = method
        else:
            print(f"INVALID METHOD USED :{method}: METHOD IS NOW SET TO 'layer_wise'")
            self.method = "layer_wise"

    def __call__(self, trainresult, model_conf):
        self.model      = trainresult.model
        self.model_name = trainresult.name
        if self.model is None :
            raise NotImplementedError()
        else:
            self.param_dict()
        
        if self.method == "layer_wise":
            self.distrib_layerwise()
        elif self.method == "network_wise":
            raise NotImplementedError()
            # self.distrib_networkwise() # [Deprecated]
        else:
            print(f"Method ERROR : {self.method=}. Must be layer_wise or network_wize")
            
        return trainresult, model_conf
        
    
    def param_dict(self):
        """
        build a dictionary of parameters dictionary sorted by layer, exemple:

            :return: {'weight'   : { 'layer1':weight_tensor1,'layer2':weight_tensor2 },

                     {'bias'     : { 'layer1':bias_tensor1, 'layer2':bias_tensor2 },

                     {'Other'    : { 'layer3':weight_tensor3}}
        """
        for name, param in self.model.named_parameters():
            layer_name, param_name = name.split(".")[0:-1],name.split(".")[-1]
            print(f"{name=} {layer_name=} {param_name=}")
            
            key_list = list(self.params.keys())
            if param_name in key_list:
                self.params[param_name]['.'.join(layer_name)]= param
            else:
                self.params[param_name] = {'.'.join(layer_name): param}
    
    def distrib_networkwise(self):
        """ Deprecated
        """
        pass
    
    def distrib_layerwise(self):
        """ Generate distribution savefiles that could be read to generate the PDF
        
            WIP - Working only with models parameters (not with activation)
        """
        # Generate files to save layers distribution informations 
        # Path : out/distribution/{network_name}
        output_model_dict   = dict()
        for param_name in list(self.params.keys()):
            layer_dict          = self.params[param_name]

            for layer_name in list(layer_dict.keys()):
                vector1d = torch.flatten(layer_dict[layer_name])
                container1d = vector1d

                min, max    = self.min_max(container1d)

                hist = torch.histc(container1d, bins = self.bins, min=min, max=max)   
                # remove all '.' in layer_name
                layer_name = layer_name.replace('.','')
                # if the layer name is ending with a number, rewrite it in format 00
                matches = re.findall(r"(\D+)(\d+)", layer_name)

                if matches:
                    layer_name = ""
                    for match in matches:
                        layer_name = f"{layer_name}{match[0]}{int(match[1]):02d}" 
                        

                print(f"{layer_name}:{param_name} min:{container1d.min().item()} - max:{container1d.max().item()} - mean:{container1d.mean().item()} - std:{container1d.std().item()}")

                output_model_dict[f"{layer_name}-{param_name}"] =[container1d.min().item(),container1d.max().item(),
                                                            container1d.mean().item(),container1d.std().item(),hist]
        
            # End for layers
        # end for model
        if self.output_layer:
            path = Path(f'{self.output_path}')/f'{self.model_name}'
            if not os.path.exists(path):
                os.makedirs(path)

            path = Path(f'{path}')/f'{self.model_name}_distribution.pkl'
            print(f"Saving in {path} ------------------------")
            f = open(path,"wb")
            pickle.dump(output_model_dict,f)
            f.close()
            
        if self.output_pdf:
            # Work only for generated file layers
            # must be rework to work "inline" without gerating files for distribution layer
            self.pdf_layerwise(line_per_page=4)

    def min_max(self, tens):
        """ Get the min and max of a tensor.
            
            Used for the histogram generated
            
            :params tens: 'class' torch tensor
            :return: min [int]

                     max [int]             
        """

        min_min = self.min
        max_max = self.max
        print(f"{tens.min()=} - {tens.max()=}")

        min = min_min if tens.min() < min_min else tens.min()
        max = max_max if tens.max() > max_max else tens.max()

        min, max = int(math.floor(min)), int(math.ceil(max))
        if min > max:
            tmp = max
            max = min
            min = tmp

        if min == max:
            min = min - 1
            max = max + 1 
        return min, max
    
    def pdf_layerwise(self, line_per_page:int=4):
        """ Generate pdf
        """
        
        network_dict = {}

        # mypath = join(self.output_path, self.model_name)
        mypath = join(self.output_path, self.model_name)
        print(f"{mypath=} - {self.output_path=} - {self.model_name=}")
        
        pkl_path = join(mypath, f"{self.model_name}_distribution.pkl")
        # get the model pkl file
        f = open(pkl_path,'rb')
        model_pkl = pickle.load(f)
        f.close()
        # get all layers names
        layers_pkl = list(model_pkl.keys())

        params_keys = []

        for layer_pkl in layers_pkl:
            # get network information with layers information stored
            print(f"{layer_pkl=}")
            layer_name, param_name = layer_pkl.split('-')
            
            min, max, mean, std, hist = model_pkl[f"{layer_name}-{param_name}"]
            
            print(f"{layer_name}")

            # Get param_type names (weight, bias, ...)
            if param_name not in params_keys:
                params_keys.append(param_name)

            # building network dict
            if layer_name in list(network_dict.keys()):
                network_dict[layer_name][param_name] = [min, max, mean, std, hist]
            else:
                network_dict[layer_name] = {param_name : [min, max, mean, std, hist]}

        keys = list(network_dict.keys())
        # sort by alphabetic order nb: work for conv and fully connected c < f 
        keys.sort()
        print(f"Layer names : {keys}")

        # option for subplot - sorted by layer type then parameters
        pdf_path = join(mypath, f"{self.model_name}-Distribution.pdf")
        pdf_pages = PdfPages(pdf_path)
        total_of_plots = len(keys) * len(params_keys)

        nb_of_pages = int(np.ceil(len(keys)/line_per_page))

        divided_keys = divide_list(keys, nb_of_pages, line_per_page)[0]
        print(f"{divided_keys=}")
        print(f"{divided_keys[0]=}")

        for page_nb, sub_keys in enumerate(divided_keys):
            fig, axes = plt.subplots(nrows=line_per_page, ncols=len(params_keys), figsize=(12, 8))
            for l_idx, l_key in enumerate(sub_keys):

                for p_idx, p_key in enumerate(params_keys):
                    # if the histogram exist
                    try:
                        # get the histogram
                        min, max, mean, std, hist = network_dict[sub_keys[l_idx]][params_keys[p_idx]]
                        dmin, dmax    = self.min_max(torch.tensor((min,max)))
                        print(f"{dmin=} - {dmax=}")
                        x = torch.arange(start=dmin,end=dmax,step=(dmax-dmin)/len(hist)).numpy()
                        x = x[0:len(hist)]
                        print(f"{sub_keys[l_idx]}:{params_keys[p_idx]} min:{min} - max:{max} - mean:{mean} - std:{std} - dmin:{dmin} - dmax:{dmax}")
                        # Create a histogram in the current subplot
                        axes[l_idx, p_idx].bar(x, hist, align='center', width=(max-min)/self.bins, linewidth=0)
                        axes[l_idx, p_idx].set_title(f"{l_key}_{p_key}|{min:4.2}|{max:4.2}|{mean:4.2}|{std:4.2}")
                    except KeyError:
                        print(f"No {params_keys[p_idx]} in {sub_keys[l_idx]} /!\\")
                        
            # Adjust spacing between subplots
            plt.tight_layout()
            # Add the current figure to the PDF
            pdf_pages.savefig(fig)

        # Close the PDF file
        pdf_pages.close()
        print(f"pdf saved : {self.output_path}")

def divide_list(input_list, m, c):
    """ Divide a list in m sublist of c element
    
    :params m: Number of sublist
    :params c: number of element per sublist

    :return: a list of all sublist od c element
    """
    # Check if n is divisible by both m and c
    if len(input_list) % (m * c) != 0:
        raise ValueError("Input list length is not divisible by m * c")

    # Divide the input list into m sub-lists of c elements each
    divided_lists = [input_list[i:i + c] for i in range(0, len(input_list), c)]

    # Split the divided_lists into m sub-lists
    final_lists = [divided_lists[i:i + m] for i in range(0, len(divided_lists), m)]

    return final_lists





from __future__ import annotations

import ast
import logging

from qualia_core.typing import TYPE_CHECKING

from .merge_dict import merge_dict

if TYPE_CHECKING:
    from qualia_core.typing import RecursiveConfigDict

logger = logging.getLogger(__name__)

def parse_args(args: list[str]) -> RecursiveConfigDict:
    def lookup_paths_to_dict(lookup_key_value_pairs: RecursiveConfigDict) -> RecursiveConfigDict:
        result_dict: RecursiveConfigDict = {}

        for k, v in lookup_key_value_pairs.items():
            if len(k) == 1: # Last item of lookup path, assign directly
                result_dict[k[0]] = v
            else: # Visit rest of path and merge into result dict
                result_dict_element = result_dict.get(k[0], {})
                if not isinstance(result_dict_element, dict):
                    logger.error("Conflicting value type for existing result key '%s', expected dict, got: %s",
                                 k[0],
                                 type(result_dict_element))
                    raise TypeError
                result_dict[k[0]] = merge_dict(result_dict_element, lookup_paths_to_dict({k[1:]: v}), merge_lists=True)

        return result_dict

    args_key_value_pairs = {}
    for arg in args:
        if not arg.startswith('--'):
            logger.error('Unknown argument: %s', arg)
            raise ValueError
        arg_expr = arg[2:] # remove '--'
        key, value = arg_expr.split('=', 2)
        key = tuple(key.split('.'))
        args_key_value_pairs[key] = ast.literal_eval(value) # Parse a literal to its correct type

    args_dict = lookup_paths_to_dict(args_key_value_pairs)
    logger.info('Command line parameters: %s', args_dict)
    return args_dict



from torch.nn import Conv1d as Conv
from torch.nn import BatchNorm1d as BatchNorm
from torch.nn import AvgPool1d as AvgPool
from torch.nn import AdaptiveAvgPool1d as AdaptiveAvgPool
from torch.nn import MaxPool1d as MaxPool
from .quantized_layers1d import QuantizedBatchNorm1d as QuantizedBatchNorm
from .quantized_layers1d import QuantizedConv1d as QuantizedConv
from .quantized_layers1d import QuantizedMaxPool1d as QuantizedMaxPool
from .quantized_layers1d import QuantizedAvgPool1d as QuantizedAvgPool
from .quantized_layers1d import QuantizedAdaptiveAvgPool1d as QuantizedAdaptiveAvgPool
from .layers import GlobalSumPool1d as GlobalSumPool
from .layers import QuantizedGlobalSumPool1d as QuantizedGlobalSumPool

__all__ = ['Conv',
           'BatchNorm',
           'AvgPool',
           'AdaptiveAvgPool',
           'MaxPool',
           'QuantizedBatchNorm',
           'QuantizedConv',
           'QuantizedMaxPool',
           'QuantizedAvgPool',
           'QuantizedAdaptiveAvgPool',
           'GlobalSumPool',
           'QuantizedGlobalSumPool']



from unittest import TestCase
from pathlib import Path

class TestUCI_HAR(TestCase):
    def setUp(self):
        self.path = Path('data')/'UCI HAR Dataset'

    def test_uci_har_raw(self):
        import numpy as np
        from qualia_core.dataset import UCI_HAR

        dataset = UCI_HAR(path=str(self.path), variant='raw')()

        train_data = np.concatenate([np.array(a.get_raw_array()) for s in dataset.sets.train for a in s.activities])
        test_data = np.concatenate([np.array(a.get_raw_array()) for s in dataset.sets.test for a in s.activities])
        train_labels = np.concatenate([np.full(len(a.samples), a.kind) for s in dataset.sets.train for a in s.activities])
        test_labels = np.concatenate([np.full(len(a.samples), a.kind) for s in dataset.sets.test for a in s.activities])

        train_bodyaccel_x = np.loadtxt(self.path/'train'/'Inertial Signals'/'body_acc_x_train.txt')
        train_bodyaccel_y = np.loadtxt(self.path/'train'/'Inertial Signals'/'body_acc_y_train.txt')
        train_bodyaccel_z = np.loadtxt(self.path/'train'/'Inertial Signals'/'body_acc_z_train.txt')
        train_bodygyro_x = np.loadtxt(self.path/'train'/'Inertial Signals'/'body_gyro_x_train.txt')
        train_bodygyro_y = np.loadtxt(self.path/'train'/'Inertial Signals'/'body_gyro_y_train.txt')
        train_bodygyro_z = np.loadtxt(self.path/'train'/'Inertial Signals'/'body_gyro_z_train.txt')
        train_totalaccel_x = np.loadtxt(self.path/'train'/'Inertial Signals'/'total_acc_x_train.txt')
        train_totalaccel_y = np.loadtxt(self.path/'train'/'Inertial Signals'/'total_acc_y_train.txt')
        train_totalaccel_z = np.loadtxt(self.path/'train'/'Inertial Signals'/'total_acc_z_train.txt')

        test_bodyaccel_x = np.loadtxt(self.path/'test'/'Inertial Signals'/'body_acc_x_test.txt')
        test_bodyaccel_y = np.loadtxt(self.path/'test'/'Inertial Signals'/'body_acc_y_test.txt')
        test_bodyaccel_z = np.loadtxt(self.path/'test'/'Inertial Signals'/'body_acc_z_test.txt')
        test_bodygyro_x = np.loadtxt(self.path/'test'/'Inertial Signals'/'body_gyro_x_test.txt')
        test_bodygyro_y = np.loadtxt(self.path/'test'/'Inertial Signals'/'body_gyro_y_test.txt')
        test_bodygyro_z = np.loadtxt(self.path/'test'/'Inertial Signals'/'body_gyro_z_test.txt')
        test_totalaccel_x = np.loadtxt(self.path/'test'/'Inertial Signals'/'total_acc_x_test.txt')
        test_totalaccel_y = np.loadtxt(self.path/'test'/'Inertial Signals'/'total_acc_y_test.txt')
        test_totalaccel_z = np.loadtxt(self.path/'test'/'Inertial Signals'/'total_acc_z_test.txt')
            
        train_y = np.loadtxt(self.path/'train'/'y_train.txt')
        test_y = np.loadtxt(self.path/'test'/'y_test.txt')

        self.assertEqual(dataset.name, 'UCI_HAR_raw')
        self.assertEqual(len(dataset.sets.train), 21)
        self.assertEqual(len(dataset.sets.test), 9)
        
        self.assertEqual(train_data.shape, (7352, 128, 9))
        self.assertEqual(test_data.shape, (2947, 128, 9))

        self.assertTrue(np.allclose(test_data[:,:,0], test_bodyaccel_x))
        self.assertTrue(np.allclose(test_data[:,:,1], test_bodyaccel_y))
        self.assertTrue(np.allclose(test_data[:,:,2], test_bodyaccel_z))
        self.assertTrue(np.allclose(test_data[:,:,3], test_bodygyro_x))
        self.assertTrue(np.allclose(test_data[:,:,4], test_bodygyro_y))
        self.assertTrue(np.allclose(test_data[:,:,5], test_bodygyro_z))
        self.assertTrue(np.allclose(test_data[:,:,6], test_totalaccel_x))
        self.assertTrue(np.allclose(test_data[:,:,7], test_totalaccel_y))
        self.assertTrue(np.allclose(test_data[:,:,8], test_totalaccel_z))

        self.assertTrue(np.allclose(train_data[:,:,0], train_bodyaccel_x))
        self.assertTrue(np.allclose(train_data[:,:,1], train_bodyaccel_y))
        self.assertTrue(np.allclose(train_data[:,:,2], train_bodyaccel_z))
        self.assertTrue(np.allclose(train_data[:,:,3], train_bodygyro_x))
        self.assertTrue(np.allclose(train_data[:,:,4], train_bodygyro_y))
        self.assertTrue(np.allclose(train_data[:,:,5], train_bodygyro_z))
        self.assertTrue(np.allclose(train_data[:,:,6], train_totalaccel_x))
        self.assertTrue(np.allclose(train_data[:,:,7], train_totalaccel_y))
        self.assertTrue(np.allclose(train_data[:,:,8], train_totalaccel_z))

        self.assertTrue(np.array_equal(train_labels, train_y-1)) # Original file is indexed from 1



from dataclasses import asdict, astuple, dataclass, fields
from typing import Any

import numpy as np
import numpy.typing


@dataclass
class Sensor:
    @classmethod
    def fieldnames(cls) -> tuple[str, ...]:
        return tuple(f.name for f in fields(cls))

    def asdict(self) -> dict[str, Any]:
        """Warning: dataclasses.asdict() and dataclasses.astuple() are recursive and may copy."""
        return asdict(self)

    def get_raw_array(self) -> numpy.typing.NDArray[np.float32]:
        return np.array(astuple(self), dtype=np.float32)



import importlib.util
import logging

logger = logging.getLogger(__name__)

__all__: list[str] = []

if importlib.util.find_spec('tensorflow') is None:
    logger.warning('Warning: cannot find TensorFlow, Keras framework will be unavailable')
elif importlib.util.find_spec('tensorflow.keras') is None:
    logger.warning('Warning: cannot find TensorFlow.Keras, Keras framework will be unavailable')
else:
    from .Keras import Keras
    from .Keras import Keras as keras
    __all__ += ['Keras', 'keras']

if importlib.util.find_spec('torch') is None:
    logger.warning('Warning: cannot find PyTorch, PyTorch framework will be unavailable')
elif importlib.util.find_spec('pytorch_lightning') is None:
    logger.warning('Warning: cannot find PyTorch Lightning, PyTorch framework will be unavailable')
else:
    from .PyTorch import PyTorch
    from .PyTorch import PyTorch as pytorch
    __all__ += ['PyTorch', 'pytorch']



from __future__ import annotations

import logging
from abc import ABC, abstractmethod

from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.process import subprocesstee

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    from pathlib import Path  # noqa: TCH003

    from qualia_core.evaluation.Evaluator import Evaluator  # noqa: TCH001

    from .Deploy import Deploy  # noqa: TCH001

logger = logging.getLogger(__name__)

class Deployer(ABC):
    evaluator: type[Evaluator] # Suggested evaluator

    @abstractmethod
    def deploy(self, tag: str) -> Deploy | None:
        ...

    def _sections_size(self, elffile: Path, size_cmd: str, section_labels: list[str]) -> int | None:
        args = ('-A', '-d', str(elffile))
        logger.info('Running: %s %s', size_cmd, ' '.join(args))
        returncode, outputs = subprocesstee.run(size_cmd, *args)
        if returncode != 0:
            return None
        outputs = outputs[1].decode().splitlines()
        logger.info([line.split() for line in outputs])
        sections = [line for line in outputs if line and any(seclabel == line.split()[0] for seclabel in section_labels)]
        sections_size = [int(line.split()[1]) for line in sections]

        return sum(sections_size)

    def _rom_size(self, elffile: Path, size_cmd: str) -> int | None:
        return self._sections_size(elffile,
                                   size_cmd,
                                   section_labels=['.isr_vector',
                                                   '.text',
                                                   '.rodata',
                                                   '.ARM',
                                                   '.preinit_array',
                                                   '.init_array',
                                                   '.fini_array',
                                                   '.data'])

    def _ram_size(self, elffile: Path, size_cmd: str) -> int | None:
        return self._sections_size(elffile,
                                   size_cmd,
                                   section_labels=['.bss', '.data'])



import io
import csv
import logging

class CSVFormatter(logging.Formatter):
    def __init__(self):
        super().__init__()
        self.output = io.StringIO()
        self.writer = csv.writer(self.output)

    def format(self, record):
        self.writer.writerow(record.msg)
        data = self.output.getvalue()
        self.output.truncate(0)
        self.output.seek(0)
        return data.strip()



from qualia_core.datamodel import DataModel

from . import Subject


class HARDataModel(DataModel[list[Subject]]):
    sets: DataModel.Sets[list[Subject]]

    def __init__(self, sets: DataModel.Sets[list[Subject]], name: str = 'HAR') -> None:
        super().__init__(sets, name)
        self.name = name



import dataclasses
import sys

import keras  # type: ignore[import-untyped]

from qualia_core.qualia import TrainResult
from qualia_core.typing import ModelConfigDict

from .PostProcessing import PostProcessing

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class RemoveKerasSoftmax(PostProcessing[keras.Model]):
    @override
    def __call__(self, trainresult: TrainResult, model_conf: ModelConfigDict) -> tuple[TrainResult, ModelConfigDict]:
        from keras.activations import softmax  # type: ignore[import-untyped]
        from keras.layers import Activation  # type: ignore[import-untyped]

        model = trainresult.model
        assert isinstance(model.layers[-1], Activation)
        assert model.layers[-1].activation == softmax

        require_compile = model._is_compiled
        if require_compile:
            compile_params = {'loss': model.loss, 'optimizer': model.optimizer, 'metrics': model.compiled_metrics.metrics}

        # Remove softmax
        model = keras.Model(model.input, model.layers[-2].output, name=model.name)

        if require_compile:
            model.compile(**compile_params)

        model.summary()
        return dataclasses.replace(trainresult, model=model), model_conf



from __future__ import annotations

import logging
import pickle as pkl
import sys
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Final

import numpy as np
import numpy.typing

from qualia_core.datamodel import RawDataModel
from qualia_core.datamodel.RawDataModel import RawData, RawDataSets

from .RawDataset import RawDataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class CORe50(RawDataset):
    """CORe50 object recognition."""

    @dataclass
    class Info:
        path: np.int32
        session: np.int8

    Info_dtype: Final[list[tuple[str, str]]] = [('path', 'U32'), ('session', 'U8')]

    test_list: Final[list[str]] = ['s3', 's7', 's10']

    class_list: Final[dict[int, int]] = {
        1: 0,
        2: 0,
        3: 0,
        4: 0,
        5: 0,
        6: 1,
        7: 1,
        8: 1,
        9: 1,
        10: 1,
        11: 2,
        12: 2,
        13: 2,
        14: 2,
        15: 2,
        16: 3,
        17: 3,
        18: 3,
        19: 3,
        20: 3,
        21: 4,
        22: 4,
        23: 4,
        24: 4,
        25: 4,
        26: 5,
        27: 5,
        28: 5,
        29: 5,
        30: 5,
        31: 6,
        32: 6,
        33: 6,
        34: 6,
        35: 6,
        36: 7,
        37: 7,
        38: 7,
        39: 7,
        40: 7,
        41: 8,
        42: 8,
        43: 8,
        44: 8,
        45: 8,
        46: 9,
        47: 9,
        48: 9,
        49: 9,
        50: 9}

    def __init__(self,
                 path: str,
                 variant: str,
                 sessions: list[str] | None = None) -> None:
        super().__init__()
        self.__path = Path(path)
        self.__variant = variant
        self.__sessions = sessions
        self.sets.remove('valid')

    def __load(self, path: Path) -> RawDataModel:
        start = time.time()

        data = np.load(path/'core50_imgs.npz')['x']

        with (path/'paths.pkl').open('rb') as f:
            paths = pkl.load(f)

        train_x_list: list[numpy.typing.NDArray[np.float32]] = []
        train_y_list: list[int] = []
        train_info_list: list[CORe50.Info] = []
        test_x_list: list[numpy.typing.NDArray[np.float32]] = []
        test_y_list: list[int] = []
        test_info_list: list[CORe50.Info] = []
        for x, p in zip(data, paths):
            session, obj, _  = p.split('/')
            label = int(obj[1:])
            if self.__variant == 'category':
                label = self.class_list[label]
            if session in self.test_list:
                test_x_list.append(x.astype(np.float32))
                test_y_list.append(label)
                test_info_list.append(CORe50.Info(path=p, session=session))
            elif self.__sessions is None or session in self.__sessions:
                train_x_list.append(x.astype(np.float32))
                train_y_list.append(label)
                train_info_list.append(CORe50.Info(path=p, session=session))

        train_x = np.array(train_x_list)
        train_y = np.array(train_y_list)
        train_info = np.array(train_info_list, dtype=self.Info_dtype)
        test_x = np.array(test_x_list)
        test_y = np.array(test_y_list)
        test_info = np.array(test_info_list, dtype=self.Info_dtype)

        logger.info('Shapes: train_x=%s, train_y=%s, test_x=%y, test_y=%y',
                    train_x.shape,
                    train_y.shape,
                    test_x.shape,
                    test_y.shape)

        train = RawData(train_x, train_y, train_info)
        test = RawData(test_x, test_y, test_info)


        logger.info('Elapsed: %s s', time.time() - start)

        return RawDataModel(sets=RawDataSets(train=train, test=test), name=self.name)

    @override
    def __call__(self) -> RawDataModel:
        if self.__variant not in ['object', 'category']:
            raise ValueError('Only object or category variants are supported')
        return self.__load(self.__path)

    @property
    @override
    def name(self) -> str:
        return f'{self.__class__.__name__}_{self.__variant}'



import asyncio
import logging
import os
import sys
from typing import BinaryIO, Optional

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class AsyncTeeProtocol(asyncio.SubprocessProtocol):
    transport: Optional[asyncio.SubprocessTransport] = None

    def __init__(self,
                 done_future: asyncio.Future[tuple[int, dict[int, bytearray]]],
                 files: Optional[dict[int, BinaryIO]] = None) -> None:
        super().__init__()
        self.done = done_future
        self.buffers: dict[int, bytearray] = {}
        self.files = files if files is not None else {}

    @override
    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        if isinstance(transport, asyncio.SubprocessTransport):
            self.transport = transport
        else:
            logger.error('Expected transport type asyncio.SubprocessTransport, got: %s', type(asyncio.SubprocessTransport))


    @override
    def pipe_data_received(self, fd: int, data: bytes) -> None:
        _ = os.write(fd, data)
        if fd not in self.buffers:
            self.buffers[fd] = bytearray()
        self.buffers[fd].extend(data)
        if fd in self.files:
            _ = self.files[fd].write(data)

    @override
    def process_exited(self) -> None:
        if self.transport is None:
            logger.error('Transport is None')
        else:
            return_code = self.transport.get_returncode()
            if return_code is None:
                logger.error('return_code is None')
            else:
                self.done.set_result((return_code, self.buffers))



from __future__ import annotations

import logging
import sys
import time
from pathlib import Path

import numpy as np
import numpy.typing

from qualia_core.datamodel import RawDataModel
from qualia_core.datamodel.Info import Info, Info_dtype
from qualia_core.datamodel.RawDataModel import RawData, RawDataSets
from qualia_core.utils.file import DirectoryReader

from .RawDataset import RawDataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class HD(RawDataset):
    """Heidelberg Digits, raw audio ('hd_audio.tar.gz')."""

    def __init__(self,
                 path: str,
                 variant: str | None = None,
                 test_subjects: list[int] | None = None) -> None:
        super().__init__()
        self.__path = Path(path)
        self.__variant = variant
        self.__test_subjects = test_subjects
        self.__dr = DirectoryReader()
        self.sets.remove('valid')

    def __load_wave(self, recording: Path) -> numpy.typing.NDArray[np.float32]:
        import soundfile as sf
        data, _ = sf.read(str(recording))
        return data.astype(np.float32)

    def __is_test_subject(self, recording: Path, test_list: list[int] | list[str]) -> bool:
        if self.__variant == 'by-subject':
            return int(recording.stem.split('_')[1].replace('speaker-', '')) in test_list
        return recording.name in test_list

    def __load(self, path: Path, test_list: list[int] | list[str]) -> RawDataModel:
        start = time.time()

        directory = self.__dr.read(path / 'audio', ext='.flac', recursive=True)


        train_x_list: list[numpy.typing.NDArray[np.float32]] = []
        train_y_list: list[int] = []
        train_info_list: list[Info] = []
        test_x_list: list[numpy.typing.NDArray[np.float32]] = []
        test_y_list: list[int] = []
        test_info_list: list[Info] = []
        for recording in directory:
            data = self.__load_wave(recording).astype(np.float32)
            label = int(recording.stem.split('-')[-1])
            if self.__is_test_subject(recording, test_list):
                test_x_list.append(data)
                test_y_list.append(label)
                test_info_list.append(Info(subject=int(recording.stem.split('_')[1].replace('speaker-', ''))))
            else:
                train_x_list.append(data)
                train_y_list.append(label)
                train_info_list.append(Info(subject=int(recording.stem.split('_')[1].replace('speaker-', ''))))

        # Zero pad to longest sample
        max_length = max(x.shape[0] for x in train_x_list + test_x_list)
        train_x_list = [np.pad(x, ((0, max_length - x.shape[0])), 'constant', constant_values=0) for x in train_x_list]
        test_x_list = [np.pad(x, ((0, max_length - x.shape[0])), 'constant', constant_values=0) for x in test_x_list]

        train_x = np.expand_dims(np.array(train_x_list), -1) # Add channels dimension
        train_y = np.array(train_y_list)
        train_info = np.array(train_info_list, dtype=Info_dtype)
        test_x = np.expand_dims(np.array(test_x_list), -1) # Add channels dimension
        test_y = np.array(test_y_list)
        test_info = np.array(test_info_list, dtype=Info_dtype)

        logger.info('Shapes: train_x=%s, train_y=%s, test_x=%y, test_y=%y',
                    train_x.shape,
                    train_y.shape,
                    test_x.shape,
                    test_y.shape)

        train = RawData(train_x, train_y, train_info)
        test = RawData(test_x, test_y, test_info)

        logger.info('Elapsed: %s s', time.time() - start)

        return RawDataModel(sets=RawDataSets(train=train, test=test), name=self.name)

    @override
    def __call__(self) -> RawDataModel:
        test_list: list[int] | list[str] = []
        if self.__variant == 'by-subject':
            if self.__test_subjects is None:
                logger.error('test_subects required for by-subject variant')
                raise ValueError
            test_list = self.__test_subjects
        else:
            with (self.__path/'test_filenames.txt').open() as f:
                test_list = f.read().splitlines()


        return self.__load(self.__path, test_list=test_list)

    @property
    @override
    def name(self) -> str:
        if self.__variant:
            return f'{super().name}_{self.__variant}'
        return super().name



from __future__ import annotations

import logging
import sys
from datetime import datetime
from pathlib import Path
from typing import Generic, NamedTuple, TypeVar

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

error_logger = logging.getLogger(__name__)

T = TypeVar('T', bound=NamedTuple)

class Logger(Generic[T]):
    logpath = Path('logs')
    prefix = ''

    def __init__(self,
                 name: str,
                 file: Path | None = None,
                 suffix: str = '.log',
                 fields: type[T] | None = None,
                 log_fields: bool = True,
                 formatter: logging.Formatter | None = None) -> None:
        super().__init__()
        self._name = name.replace('.', '_') # module names contain dots but namedtuple requires identifier
        self.__file = file
        self.__suffix = suffix
        self.__formatter = formatter
        self.__init_fields = fields
        self.__fields: type[T] | None = None
        self.__log_fields = log_fields
        self._content: list[T] | list[tuple[str, ...]]= []
        self.logger: logging.Logger | None = None
        self.filehandler: logging.FileHandler | None = None

    def _lazy_init(self) -> None:
        """Lazy init allows setting configuring global parameters such as path/prefix even after instanciation
        (when classes with CSVLogger as class member are loaded) but before logger is called for the first time.
        """
        # Add logger to name to prevent conflict with other module-local loggers
        self.logger = logging.getLogger(f'qualia_core.logger.{self._name}')

        self.logger.propagate = False # Don't use logger hierarchy since somehow root logger may have a handler already
        self.logger.setLevel(logging.DEBUG)

        if not self.logger.hasHandlers():
            if self.__file is None:
                self.__file = Path(self._name)/f'{self.prefix}{datetime.now():%Y-%m-%d_%H-%M-%S}{self.__suffix}'
            self.__file = self.logpath/self.__file
            self.__file.parent.mkdir(parents=True, exist_ok=True)

            self.filehandler = logging.FileHandler(self.__file, delay=True)
            self.filehandler.setLevel(logging.DEBUG)
            if self.__formatter:
                self.filehandler.setFormatter(self.__formatter)
            self.logger.addHandler(self.filehandler)
        if self.__init_fields is not None:
            self.fields = self.__init_fields

    def __call__(self, data: T) -> None:
        if self.logger is None:
            self._lazy_init()
        if self.logger is None:
            error_logger.error('Could not initialize logger %s', self._name)
            raise RuntimeError

        str_args = tuple(str(v) for v in data)

        self.logger.info(str_args)

        if self.fields is not None:
            self._content.append(data)
        else:
            self._content.append(str_args)

    def __del__(self) -> None:
        if self.filehandler is not None:
            self.filehandler.close()
            if self.logger is not None:
                self.logger.removeHandler(self.filehandler)

    @override
    def __repr__(self) -> str:
        return str(self.content)

    @property
    def fields(self) -> type[T] | None:
        return self.__fields

    @fields.setter
    def fields(self, val: type[T]) -> None:
        if self.__fields is not None:
            error_logger.error("'fields' can only be assigned once for each instance of %s", type(self))
            raise AttributeError
        self.__fields = val

        # Write column names to file
        if self.logger is None:
            self._lazy_init()
        if self.logger is None:
            error_logger.error('Could not initialize logger %s', self._name)
            raise RuntimeError

        if self.__log_fields:
            self.logger.info(val._fields)

    @property
    def content(self) -> list[T] | list[tuple[str, ...]]:
        return self._content



from __future__ import annotations

import os
import sys
import typing

if sys.version_info >= (3, 12):
    from typing import NotRequired, TypeAliasType, TypedDict
else:
    from typing_extensions import NotRequired, TypeAliasType, TypedDict

TYPE_CHECKING = typing.TYPE_CHECKING or os.environ.get('SPHINX_AUTODOC', False)



if TYPE_CHECKING:
    RecursiveConfigUnion = typing.Union[dict[str, 'RecursiveConfigUnion'],
                                        list['RecursiveConfigUnion'],
                                        str,
                                        int,
                                        float,
                                        bool]
else:
    RecursiveConfigUnion = TypeAliasType('RecursiveConfigUnion', typing.Union[dict[str, 'RecursiveConfigUnion'],
                                        list['RecursiveConfigUnion'],
                                        str,
                                        int,
                                        float,
                                        bool])
RecursiveConfigDict = dict[str, 'RecursiveConfigUnion']


class GenericModuleConfigDict(TypedDict):
    kind: str
    params: NotRequired[RecursiveConfigDict]

class BenchConfigDict(TypedDict):
    name: str
    seed: int
    first_run: int
    last_run: int
    plugins: NotRequired[list[str]]

class ExperimentTrackingConfigDict(GenericModuleConfigDict):
    ...

class LearningFrameworkConfigDict(GenericModuleConfigDict):
    ...

class ConverterConfigDict(GenericModuleConfigDict):
    ...

class DeployerConfigDict(TypedDict):
    kind: NotRequired[str] # Not mandatory since deployer can be suggested by converter.
    params: NotRequired[RecursiveConfigDict]

class EvaluatorConfigDict(TypedDict):
    kind: NotRequired[str] # Not mandatory since evaluator can be suggested by deployer
    params: NotRequired[RecursiveConfigDict]

class DeployConfigDict(TypedDict):
    target: str
    converter: ConverterConfigDict
    deployer: NotRequired[DeployerConfigDict]
    evaluator: NotRequired[EvaluatorConfigDict]
    quantize: list[str]
    optimize: NotRequired[list[str]]
    compress: NotRequired[list[int]]
    limit: NotRequired[int]

class DatasetConfigDict(GenericModuleConfigDict):
    ...

class PreprocessingConfigDict(GenericModuleConfigDict):
    ...

class DataAugmentationConfigDict(GenericModuleConfigDict):
    ...

class PostprocessingConfigDict(GenericModuleConfigDict):
    export: NotRequired[bool]

class SchedulerConfigDict(GenericModuleConfigDict):
    ...

class OptimizerConfigDict(GenericModuleConfigDict):
    scheduler: NotRequired[SchedulerConfigDict]

class QuantizerConfigDict(TypedDict):
    quant_enable: NotRequired[bool]
    LSQ: NotRequired[bool]
    bits: NotRequired[int]
    force_q: NotRequired[int]
    range_setting: NotRequired[str]
    roundtype: NotRequired[str]
    tensor_type: NotRequired[str]
    quantype: NotRequired[str]
    is_asymmetric: NotRequired[bool]

class QuantizationConfigDict(QuantizerConfigDict):
    act: NotRequired[QuantizerConfigDict]
    input: NotRequired[QuantizerConfigDict]  # noqa: A003
    w: NotRequired[QuantizerConfigDict]
    bias: NotRequired[QuantizerConfigDict]

class ModelParamsConfigDict(TypedDict):
    input_shape: typing.Union[list[int], tuple[int, ...]]
    output_shape: typing.Union[list[int], tuple[int, ...]]
    quant_params: NotRequired[QuantizationConfigDict]

class UpdateConfigDict(TypedDict):
    quant_params: NotRequired[QuantizationConfigDict]

class ModelCommonConfigDict(TypedDict):
    params: NotRequired[typing.Union[RecursiveConfigDict, ModelParamsConfigDict]]
    epochs: NotRequired[int]
    batch_size: NotRequired[int]
    load: NotRequired[bool]
    train: NotRequired[bool]
    evaluate: NotRequired[bool]
    optimizer: NotRequired[OptimizerConfigDict]
    update: NotRequired[UpdateConfigDict]

class ModelTemplateConfigDict(ModelCommonConfigDict):
    kind: NotRequired[str]

class ModelConfigDict(ModelCommonConfigDict):
    kind: str
    name: str
    disabled: NotRequired[bool]

class ConfigDict(TypedDict):
    bench: BenchConfigDict
    experimenttracking: NotRequired[ExperimentTrackingConfigDict]
    learningframework: LearningFrameworkConfigDict
    deploy: NotRequired[DeployConfigDict]
    dataset: DatasetConfigDict

    preprocessing: NotRequired[list[PreprocessingConfigDict]]
    data_augmentation: NotRequired[list[GenericModuleConfigDict]]
    postprocessing: NotRequired[list[PostprocessingConfigDict]]
    model: list[ModelConfigDict]
    model_template: ModelTemplateConfigDict

__all__ = ['TYPE_CHECKING', 'RecursiveConfigDict']



from collections import OrderedDict
from typing import cast

from qualia_core.learningmodel.pytorch.layers import QuantizedAdd
from qualia_core.learningmodel.pytorch.layers.quantized_layers import QuantizedIdentity, QuantizedLinear, QuantizedReLU
from qualia_core.learningmodel.pytorch.ResNet import BasicBlock, ResNet
from torch import nn


class QuantizedBasicBlock(BasicBlock):
    def __init__(self,
                    layers_t,
                    in_planes,
                    planes,
                    kernel_size,
                    stride,
                    padding,
                    batch_norm,
                    bn_momentum,
                    force_projection_with_stride,
                    quantize_add,
                    quant_params,
                    fused_relu=True):
        super().__init__(layers_t, in_planes, planes, kernel_size, stride, padding, batch_norm, bn_momentum, force_projection_with_stride)
        self.fused_relu = fused_relu

        
        
        
        
        self.conv1 = layers_t.QuantizedConv(in_planes,
                                            planes,
                                            kernel_size=kernel_size,
                                            stride=1,
                                            padding=padding,
                                            bias=not batch_norm,
                                            quant_params=quant_params,
                                            activation=nn.ReLU() if fused_relu and self.stride == 1 and not batch_norm else None)
        if batch_norm:
            self.bn1 = layers_t.QuantizedBatchNorm(planes,
                                                   momentum=bn_momentum,
                                                   quant_params=quant_params,
                                                   activation=nn.ReLU() if fused_relu and self.stride == 1 else None)


        if self.stride != 1:
            self.pool1 = layers_t.QuantizedMaxPool(stride, quant_params=quant_params, activation=nn.ReLU() if fused_relu else None)
        if fused_relu:
            del self.relu1
        else:
            self.relu1 = QuantizedReLU(quant_params=quant_params)


        self.conv2 = layers_t.QuantizedConv(planes,
                                            planes,
                                            kernel_size=kernel_size,
                                            stride=1,
                                            padding=padding,
                                            bias=not batch_norm,
                                            quant_params=quant_params)

        if batch_norm:
            self.bn2 = layers_t.QuantizedBatchNorm(planes, momentum=bn_momentum, quant_params=quant_params)


        if self.stride != 1:
            self.smax = layers_t.QuantizedMaxPool(stride, quant_params=quant_params)
        if self.in_planes != self.expansion*self.planes or force_projection_with_stride and self.stride != 1:
            self.sconv = layers_t.QuantizedConv(in_planes, self.expansion*planes, kernel_size=1, stride=1, bias=not batch_norm, quant_params=quant_params)
            if batch_norm:
                self.sbn = layers_t.QuantizedBatchNorm(
                        self.expansion*planes, momentum=bn_momentum, quant_params=quant_params)

        self.add = QuantizedAdd(quant_params=quant_params , activation=nn.ReLU() if fused_relu else None)
        if fused_relu:
            del self.relu
        else:
            self.relu = QuantizedReLU(quant_params=quant_params)
            
    def forward(self, x):

        out = self.conv1(x)

        if self.batch_norm:
            out = self.bn1(out)
        if self.stride != 1:
            out = self.pool1(out)
        if not self.fused_relu:
            out = self.relu1(out)

        out = self.conv2(out)
        if self.batch_norm:
            out = self.bn2(out)


        # shortcut
        tmp = x

        if self.in_planes != self.expansion*self.planes or self.force_projection_with_stride and self.stride != 1:
            tmp = self.sconv(tmp)

            if self.batch_norm:
                tmp = self.sbn(tmp)

        if self.stride != 1:
            tmp = self.smax(tmp)

        out = self.add(out, tmp)
        if not self.fused_relu:
            out = self.relu(out)
        return out

class QuantizedResNet(ResNet):
    def __init__(self,
        input_shape,
        output_shape,
        filters: tuple=(15, 15, 30, 60, 120),
        kernel_sizes: tuple=(3, 3, 3, 3, 3),

        num_blocks: tuple=(2, 2, 2, 2),
        strides: tuple=(1, 1, 2, 2, 2),
        paddings: int=(1, 1, 1, 1, 1),
        prepool: int=1,
        postpool: str='max',
        batch_norm: bool=False,
        bn_momentum: float=0.1,
        force_projection_with_stride: bool=True,
        dims: int=1,
        
        quantize_linear: bool=True,
        quantize_add: bool=True,
        fused_relu: bool=True,
        quant_params: dict = None):

        #if batch_norm:
        #    raise ValueError('BatchNorm unsupported in quantized Resnet')

        #if prepool > 1:
        #    raise ValueError('AvgPool unsupported in quantized Resnet')
        
        if "bits" in quant_params :
            quant_params["bits"] = int(quant_params["bits"]) # Force conversion from TOML int to plain Python int
            if quant_params["bits"] < 1:
                raise ValueError('bits must be set to a strictly positive integer')
        else :
            raise ValueError('bits must exist in quant_params conf')

        if dims == 1:
            import qualia_core.learningmodel.pytorch.layers1d as layers_t
        elif dims == 2:
            import qualia_core.learningmodel.pytorch.layers2d as layers_t
        else:
            raise ValueError('Only dims=1 or dims=2 supported')

        super().__init__(input_shape=input_shape,
                         output_shape=output_shape,
                         filters=filters,
                         kernel_sizes=kernel_sizes,
                         num_blocks=num_blocks,
                         strides=strides,
                         paddings=paddings,
                         prepool=prepool,
                         postpool=postpool,
                         batch_norm=batch_norm,
                         bn_momentum=bn_momentum,
                         force_projection_with_stride=force_projection_with_stride,
                         dims=dims,
                         basicblockbuilder=lambda in_planes, planes, kernel_size, stride, padding: QuantizedBasicBlock(
                                                                                layers_t=layers_t,
                                                                                in_planes=in_planes,
                                                                                planes=planes,
                                                                                kernel_size=kernel_size,
                                                                                stride=stride,
                                                                                padding=padding,
                                                                                batch_norm=batch_norm,
                                                                                bn_momentum=bn_momentum,
                                                                                force_projection_with_stride=force_projection_with_stride,
                                                                                quantize_add=quantize_add,
                                                                                quant_params=quant_params,
                                                                                fused_relu=fused_relu)
                         )
        self.fused_relu = fused_relu
        self.batch_norm = batch_norm

        self.identity1 = QuantizedIdentity(quant_params=quant_params)
        # Make sure identity1 appears before all of parent ResNet modules to keep a more sensible iterate/print order
        # self._modules is assumed to be an OrderedDict which is the case in PyTorch 2.0
        cast(OrderedDict[str, nn.Module], self._modules).move_to_end('identity1', last=False)

        if prepool > 1:
            self.prepool = layers_t.QuantizedAvgPool(prepool, bits=bits, force_q=force_q)

        self.conv1 = layers_t.QuantizedConv(input_shape[-1],
                                            filters[0],
                                            kernel_size=kernel_sizes[0],
                                            stride=strides[0],
                                            padding=paddings[0],
                                            bias=not batch_norm,
                                            quant_params=quant_params,
                                            activation=nn.ReLU() if fused_relu and not batch_norm else None)
        if self.batch_norm:
            self.bn1 = layers_t.QuantizedBatchNorm(self.in_planes,
                                                   momentum=bn_momentum,
                                                   quant_params=quant_params,
                                                   activation=nn.ReLU() if fused_relu else None)
        if fused_relu:
            del self.relu1
        else:
            self.relu1 = QuantizedReLU(quant_params=quant_params)

        if postpool == 'avg':
            self.postpool = layers_t.QuantizedAdaptiveAvgPool(1, quant_params=quant_params)
        elif postpool == 'max':
            self.postpool = layers_t.QuantizedMaxPool(tuple(self._fm_dims), quant_params=quant_params)
        
        self.flatten = nn.Flatten()
        # Replace Linear with the quantized variant
        if quantize_linear:
            self.linear = QuantizedLinear(self.in_planes*QuantizedBasicBlock.expansion, output_shape[0], quant_params=quant_params)

    def forward(self, x):

        out = self.identity1(x)
        
        if hasattr(self, 'prepool'):
            out = self.prepool(out)

        out = self.conv1(out)
        if self.batch_norm:
            out = self.bn1(out)
        if not self.fused_relu:
            out = self.relu1(out)

        for i in range(len(self.layers)):
            out = self.layers[i](out)

        if hasattr(self, 'postpool'):
            out = self.postpool(out)

        out = self.flatten(out)
        out = self.linear(out)
        return out



from unittest import TestCase

class TestModelsQuantizedResnet2D(TestCase):
    def setUp(self):
        import numpy as np
        from qualia_core.datamodel import RawDataModel
        from qualia_core.datamodel.RawDataModel import RawData, RawDataSets
        train = RawData(np.ones((2, 64, 64, 1), dtype=np.float32), np.array([[1, 0], [1, 0]]))
        test = RawData(np.ones((2, 64, 64, 1), dtype=np.float32), np.array([[1, 0], [0, 1]]))
        self.__data = RawDataModel(RawDataSets(train=train, test=test), name='test_quantizedresnet2d')
        self.__model_params = {
            'filters': (4, 6),
            'kernel_sizes': (5, 5),

            'num_blocks': (2,),
            'strides': (1, 2),
            'paddings': (2, 2),

            'prepool': 1,
            'batch_norm': False,
            'bn_momentum': 0.95,

            'dims': 2,

            'quant_params': {'bits': 8,
                             'quantype': 'fxp',
                             'roundtype': 'floor',
                             'range_setting': 'minmax',
                             'LSQ': False},
            'fused_relu': True,
            }

    def test_quantized_resnet_2d_pytorch(self):
        from qualia_core import qualia
        from qualia_core.learningframework import PyTorch
        from qualia_core.learningmodel.pytorch import QuantizedResNet
        from qualia_core.learningmodel.pytorch.layers import QuantizedAdd
        from qualia_core.learningmodel.pytorch.quantized_layers2d import QuantizedConv2d, QuantizedMaxPool2d
        from qualia_core.learningmodel.pytorch.layers.quantized_layers import QuantizedIdentity, QuantizedLinear
        from torch.nn import Flatten, Linear, ReLU

        model = QuantizedResNet

        framework = PyTorch(enable_progress_bar=False)

        trainresult = qualia.train(self.__data,
                        train_epochs=1,
                        iteration=1,
                        model_name='test_quantized_resnet_2d_pytorch',
                        model=model,
                        model_params=self.__model_params,
                        optimizer={'kind': 'Adam'},
                        framework=framework,
                        )
        self.assertEqual(trainresult.name, 'test_quantized_resnet_2d_pytorch')
        self.assertEqual(trainresult.i, 1)
        #self.assertEqual(trainresult.model.input_shape, (None, 1, 1)) # Not supported in PyTorch
        #self.assertEqual(trainresult.model.output_shape, (None, 2)), Not supported in PyTorch

        self.assertEqual(len(list(trainresult.model.children())), 6)
        self.assertEqual(len(list(trainresult.model.layers.children())), 1)
        self.assertEqual(len(list(trainresult.model.layers[0].children())), 2)
        self.assertEqual(len(list(trainresult.model.layers[0][0].children())), 6)
        self.assertEqual(len(list(trainresult.model.layers[0][1].children())), 3)

        # input quantization
        self.assertEqual(type(trainresult.model.identity1), QuantizedIdentity)

        # prepool
        self.assertFalse(hasattr(trainresult.model, 'prepool'))

        # preconv
        self.assertEqual(type(trainresult.model.conv1), QuantizedConv2d)
        self.assertEqual(tuple(trainresult.model.conv1.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.conv1.weight.shape), (4, 1, 5, 5)) # 4*1*5*5 = 100
        self.assertEqual(tuple(trainresult.model.conv1.bias.shape), (4, )) # 4
        self.assertFalse(hasattr(trainresult.model, 'bn1'))
        self.assertEqual(type(trainresult.model.conv1.activation), ReLU)

        # 1st res block
        # 1st conv
        self.assertEqual(type(trainresult.model.layers[0][0].conv1), QuantizedConv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv1.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv1.weight.shape), (6, 4, 5, 5)) # 6*4*5*5 = 600
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv1.bias.shape), (6, )) # 6
        self.assertFalse(hasattr(trainresult.model.layers[0][0], 'bn1'))
        self.assertEqual(type(trainresult.model.layers[0][0].pool1), QuantizedMaxPool2d)
        self.assertEqual(trainresult.model.layers[0][0].pool1.kernel_size, 2)
        self.assertEqual(trainresult.model.layers[0][0].pool1.stride, 2)
        self.assertEqual(type(trainresult.model.layers[0][0].pool1.activation), ReLU)
        # 2nd conv
        self.assertEqual(type(trainresult.model.layers[0][0].conv2), QuantizedConv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv2.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv2.weight.shape), (6, 6, 5, 5)) # 6*6*5*5 = 900
        self.assertEqual(tuple(trainresult.model.layers[0][0].conv2.bias.shape), (6, )) # 6
        self.assertFalse(hasattr(trainresult.model.layers[0][0], 'bn2'))
        # shortcut
        self.assertEqual(type(trainresult.model.layers[0][0].sconv), QuantizedConv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][0].sconv.padding), (0, 0))
        self.assertEqual(tuple(trainresult.model.layers[0][0].sconv.weight.shape), (6, 4, 1, 1)) # 6*4*1*1 = 24
        self.assertEqual(tuple(trainresult.model.layers[0][0].sconv.bias.shape), (6, )) # 6
        self.assertFalse(hasattr(trainresult.model.layers[0][0], 'sbn'))
        self.assertEqual(type(trainresult.model.layers[0][0].smax), QuantizedMaxPool2d)
        self.assertEqual(trainresult.model.layers[0][0].smax.kernel_size, 2)
        self.assertEqual(trainresult.model.layers[0][0].smax.stride, 2)
        # add
        self.assertEqual(type(trainresult.model.layers[0][0].add), QuantizedAdd)
        self.assertEqual(type(trainresult.model.layers[0][0].add.activation), ReLU)


        # 2nd conv block
        # 1st conv
        self.assertEqual(type(trainresult.model.layers[0][1].conv1), QuantizedConv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv1.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv1.weight.shape), (6, 6, 5, 5)) # 6*6*5*5 = 900
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv1.bias.shape), (6, )) # 6
        self.assertFalse(hasattr(trainresult.model.layers[0][1], 'bn1'))
        self.assertEqual(type(trainresult.model.layers[0][1].conv1.activation), ReLU)
        # 2nd conv
        self.assertEqual(type(trainresult.model.layers[0][1].conv2), QuantizedConv2d)
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv2.padding), (2, 2))
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv2.weight.shape), (6, 6, 5, 5)) # 6*6*5*5 = 900
        self.assertEqual(tuple(trainresult.model.layers[0][1].conv2.bias.shape), (6, )) # 6
        self.assertFalse(hasattr(trainresult.model.layers[0][1], 'bn2'))
        # add
        self.assertEqual(type(trainresult.model.layers[0][1].add), QuantizedAdd)
        self.assertEqual(type(trainresult.model.layers[0][1].add.activation), ReLU)

        self.assertEqual(type(trainresult.model.postpool), QuantizedMaxPool2d)
        self.assertEqual(tuple(trainresult.model.postpool.kernel_size), (32, 32))
        self.assertEqual(tuple(trainresult.model.postpool.stride), (32, 32))

        self.assertEqual(type(trainresult.model.flatten), Flatten)

        # fc
        self.assertEqual(type(trainresult.model.linear), QuantizedLinear)
        self.assertEqual(trainresult.model.linear.weight.shape, (2, 6)) # 2*6 = 12
        self.assertEqual(trainresult.model.linear.bias.shape, (2, )) # 2


        self.assertEqual(trainresult.mem_params, (5*5*1*4 + 4 + 5*5*4*6 + 6 + 5*5*6*6 + 6 + 1*1*4*6 + 6 + 5*5*6*6 + 6 + 5*5*6*6 + 6 + 6*2 + 2) * 4)
        self.assertEqual(trainresult.acc, 0.5) # Same data in one or the other class, should have 50% acc
        self.assertEqual(trainresult.framework, framework)



from pathlib import Path


class TestCIFAR10:
    path = Path('data')/'cifar-10-batches-py'

    def test_cifar10(self) -> None:
        import numpy as np
        from qualia_core.dataset import CIFAR10

        dataset = CIFAR10(path=str(self.path))()

        import pickle
        with (self.path/'test_batch').open('rb') as fo:
            raw = pickle.load(fo, encoding='bytes')
            test_y = np.array(raw[b'labels'])

        assert dataset.sets.train.x.shape == (50000, 32, 32, 3)
        assert dataset.sets.train.x.dtype == np.float32
        assert dataset.sets.test.x.shape == (10000, 32, 32, 3)
        assert dataset.sets.test.x.dtype == np.float32
        assert dataset.sets.train.y.shape == (50000, )
        assert dataset.sets.test.y.shape == (10000, )

        assert np.array_equal(dataset.sets.test.y, test_y)



from __future__ import annotations

import csv
from dataclasses import dataclass
from types import SimpleNamespace
from typing import TYPE_CHECKING, Callable, Generic, TypeVar

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path

T = TypeVar('T')
U = TypeVar('U')

class CSVReader(Generic[T]):
    @dataclass
    class CSVResult(Generic[U]):
        filename: Path
        content: list[U]

    def read_callback(self,
                      filename: Path,
                      labels: type[T],
                      callback: Callable[[Generator[T, None, None]], None],
                      delimiter: str = ',',
                      skip_header: bool = True) -> None:
        with filename.open(newline='') as f:
            reader = csv.reader(f, delimiter=delimiter)

            if skip_header:
                _ = next(reader)

            callback(labels(*row) for row in reader)

    def read(self,
             filename: Path,
             labels: type[T],
             delimiter: str = ',',
             skip_header: bool = True) -> CSVResult[T]:
        full_content = SimpleNamespace(val=[])

        def callback(content: Generator[T, None, None]) -> None:
            full_content.val = list(content)

        self.read_callback(filename=filename,
                           callback=callback,
                           delimiter=delimiter,
                           labels=labels,
                           skip_header=skip_header)
        return CSVReader.CSVResult(filename=filename, content=full_content.val)



import numpy as np

from .Preprocessing import Preprocessing

class Class2BinMatrix(Preprocessing):
    '''Warning: must be applied after Window'''
    def __init__(self, classes: int=None):
        self.__classes = classes

    def __call__(self, datamodel):
        for _, s in datamodel:
            if len(s.y.shape) != 1:
                raise ValueError(f'Unsupported dimensions for {self.__class__.__name__}: {len(s.y.shape)}')
            if len(s.y) <= 0: # Handle empty sets
                continue

            if not self.__classes:
                s.y = np.eye(np.max(s.y) + 1)[s.y]
            else:
                s.y = np.eye(self.__classes)[s.y]

        return datamodel



from __future__ import annotations

import logging
import statistics
import sys
import time
from datetime import datetime
from pathlib import Path
from typing import Any, NamedTuple

from qualia_core.evaluation.Evaluator import Evaluator
from qualia_core.evaluation.Stats import Stats
from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.logger.CSVLogger import CSVLogger

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation  # noqa: TCH001
    from qualia_core.datamodel.RawDataModel import RawDataModel  # noqa: TCH001
    from qualia_core.learningframework.LearningFramework import LearningFramework  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class Result(NamedTuple):
    i: int = -1
    y: int = -1
    score: float = -1
    time: float = -1


class Qualia(Evaluator):
    """Custom evaluation loop for Qualia embedded implementations like TFLite Micro and Qualia-CodeGen."""

    def __init__(self,
                 dev: str = '/dev/ttyUSB0',
                 baudrate: int = 921600,
                 timeout: int = 30, # 30 seconds default transmission timeout
                 shuffle: bool = False) -> None:  # noqa: FBT001, FBT002
        super().__init__()

        self.__dev = Path(dev)
        self.__baudrate = baudrate
        self.__shuffle = shuffle
        self.__timeout = timeout

    def __get_dev(self, path: Path) -> Path | None:
        dev: Path | None = None
        logger.info('Waiting for device "%s"…', path)
        start_time = time.time()
        while dev is None:
            if time.time() - start_time > self.__timeout:
                logger.error('Timeout looking up for device "%s"', path)
                break
            devs = list(path.parent.glob(path.name))
            if devs:
                if len(devs) > 1:
                    logger.warning('%d devices matched, using first device %s', len(devs), devs[0])
                dev = devs[0]
            time.sleep(0.1)
        return dev

    @override
    def evaluate(self,
                 framework: LearningFramework[Any],
                 model_kind: str,
                 dataset: RawDataModel,
                 target: str,
                 tag: str,
                 limit: int | None = None,
                 dataaugmentations: list[DataAugmentation] | None = None) -> Stats | None:
        import serial

        dev = self.__get_dev(self.__dev)
        if dev is None:
            logger.error('No device found.')
            return None

        if dataset.sets.test is None:
            logger.error('Test dataset is required')
            raise ValueError

        test_x = dataset.sets.test.x.copy()
        test_y = dataset.sets.test.y.copy()

        # Shuffle
        if self.__shuffle:
            test_x, test_y = self.shuffle_dataset(test_x, test_y)

        test_x, test_y = self.apply_dataaugmentations(framework, dataaugmentations, test_x, test_y)

        test_x, test_y = self.limit_dataset(test_x, test_y, limit)

        logger.info('Reset the target…')
        s = serial.Serial(str(dev), self.__baudrate, timeout=self.__timeout)
        logger.info('Device: %s', s.name)
        r = s.readline().decode('cp437')

        start_time = time.time()
        logger.info('Waiting for READY from device "%s"…', dev)
        while 'READY' not in r:
            if time.time() - start_time > self.__timeout:
                logger.error('Timeout waiting for READY for device "%s"', dev)
                return None

            r = s.readline().decode('cp437')
            time.sleep(0.1)

        # create log directory
        (Path('logs')/'evaluate'/target).mkdir(parents=True, exist_ok=True)

        results: list[Result] = [] # read from target

        log: CSVLogger[Result] = CSVLogger(name=__name__,
                        file=Path('evaluate')/target/f'{tag}_{datetime.now():%Y-%m-%d_%H-%M-%S}.txt')  # noqa: DTZ005 system tz ok
        log.fields = Result

        for i, line in enumerate(test_x):
            msg = ','.join(map(str, line.flatten())) + '\r\n'
            _ = s.write(msg.encode('cp437')) # Send test vector

            r = s.readline() # Read acknowledge
            tstart = time.time() # Start timer
            r = r.decode('cp437')
            if not r or int(r) != len(msg): # Timed out or didn't receive all the data
                logger.error('%d: Transmission error: %s != %d', i, int(r), len(msg))
                return None

            r = s.readline() # Read result
            tstop = time.time() # Stop timer
            r = r.decode('cp437').rstrip().split(',')
            r = Result(i=int(r[0]),
                       y=int(r[1]),
                       score=float(r[2]),
                       time=tstop-tstart)
            logger.info('%d: %s', i, str(r))
            results.append(r)

            # Log result to file
            log(r)

        avg_time = statistics.mean([r.time for r in results])

        return Stats(avg_time=avg_time, accuracy=self.compute_accuracy(preds=[r.y for r in results], truth=test_y))

        # avg it/secs
        # ram usage
        # rom usage
        # cpu type
        # cpu model
        # cpu freq
        # accuracy
        # power consumption



from __future__ import annotations

from collections import namedtuple
from importlib.resources import files
from pathlib import Path

from qualia_core.deployment.Deployer import Deployer
from qualia_core.utils.path import resources_to_path
from qualia_core.utils.process import subprocesstee


class Linux(Deployer):
    import qualia_core.evaluation.host.Qualia as evaluator  # Suggested evaluator

    def __init__(self,
                 cxxflags: list[str] | None = None,
                 modeldir: str | Path | None = None,
                 projectdir: str | Path | None = None,
                 outdir: str | Path | None = None) -> None:
        super().__init__()
        self.__cxxflags = cxxflags if cxxflags is not None else ['-ftrapv', '-Wall', '-Wextra', '-Werror=double-promotion',
                                                                 '-pedantic', '-Ofast', '-ffunction-sections', '-fdata-sections',
                                                                 '-fgraphite-identity', '-floop-nest-optimize',
                                                                 '-floop-parallelize-all', '-std=c++17']
        self.__modeldir = Path(modeldir) if modeldir is not None else Path('out')/'qualia_codegen'
        self.__projectdir = (Path(projectdir) if projectdir is not None
                             else resources_to_path(files('qualia_codegen_core.examples'))/'Linux')
        self.__outdir = Path(outdir) if outdir is not None else Path('out')/'deploy'/'Linux'

    def __run(self, cmd, *args):
        print(cmd, *args)
        returncode, outputs = subprocesstee.run(str(cmd), *args)
        return returncode == 0

    def __create_outdir(self):
        self.__outdir.mkdir(parents=True, exist_ok=True)

    def __build(self, tag: str, model):
        modeldir = self.__modeldir/model.name
        return self.__run('g++',
                          '-o', str(self.__outdir/f'{tag}_Linux'),
                           str(self.__projectdir/'main.cpp'),
                           str(modeldir/'model.c'),
                           f'-I{modeldir}',
                           f'-I{modeldir}/include',
                          *self.__cxxflags,
                        )

    def prepare(self, tag, model, optimize: str, compression: int):
        if optimize:
            raise ValueError(f'No optimization available for {self.__class__.__name__}')
        if compression != 1:
            raise ValueError(f'No compression available for {self.__class__.__name__}')

        self.__create_outdir()

        if not self.__build(tag=tag, model=model):
            return None
        return self

    def deploy(self, tag):
        print(self.__class__.__name__, 'Info: running locally, nothing to deploy')

        return namedtuple('Deploy', ['rom_size', 'ram_size', 'evaluator'])(self._rom_size(tag), self._ram_size(tag), self.evaluator)

    def _rom_size(self, tag):
        return super()._rom_size(self.__outdir/f'{tag}_Linux', 'size')

    def _ram_size(self, tag: str):
        return super()._ram_size(self.__outdir/f'{tag}_Linux', 'size')



from __future__ import annotations

import shutil
import sys
from pathlib import Path

from qualia_core.typing import TYPE_CHECKING

from .Eclipse import Eclipse

if TYPE_CHECKING:
    from qualia_core.deployment.Deploy import Deploy  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class NucleiStudio(Eclipse):
    def __init__(self,
                 projectname: str,
                 projectdir: Path,
                 outdir: Path | None = None) -> None:
        outdir = outdir if outdir is not None else Path('out')/'deploy'/'NucleiStudio'

        nuclei_dir = Path('/opt')/'nuclei'
        nucleistudio_bin = nuclei_dir/'NucleiStudio'/'NucleiStudio'

        dfu_util_bin = Path('/usr')/'bin'/'dfu-util'

        riscv_size_bin = nuclei_dir/'gcc'/'bin'/'riscv-nuclei-elf-size'

        super().__init__(eclipse_bin=nucleistudio_bin,
                         size_bin=riscv_size_bin,
                         upload_bin=dfu_util_bin,
                         projectname=projectname,
                         projectdir=projectdir,
                         outdir=outdir)

    @override
    def _copy_buildproduct(self, tag: str) -> None:
        shutil.copy(self._projectdir/self._buildtype/f'{self._projectname}.bin', self._outdir/f'{tag}.bin')
        return super()._copy_buildproduct(tag=tag)

    @override
    def _upload(self,
                tag: str,
                logdir: Path,
                args: tuple[str, ...] | None = None,
                cmd: Path | None = None) -> bool:
        args = ('-s', '0x08000000:leave', '-D', str(self._outdir/f'{tag}.bin'))
        return super()._upload(tag=tag, logdir=logdir, args=args)

    @override
    def deploy(self, tag: str) -> Deploy | None:
        _ = input('Put target in programming mode and press Enter…')
        return super().deploy(tag=tag)
